import json
import logging
import os
import time
import dateutil.parser
from typing import Any, Dict, Optional, Tuple, Type, Union
from urllib.parse import urljoin

import requests  # type: ignore
from pydantic import BaseModel, field_validator, Field, model_validator
from requests import RequestException
from prometrix.connect.aws_connect import AWSPrometheusConnect
from prometrix.models.prometheus_config import PrometheusConfig as BasePrometheusConfig
from holmes.core.tools import (
    CallablePrerequisite,
    StructuredToolResult,
    Tool,
    ToolInvokeContext,
    ToolParameter,
    StructuredToolResultStatus,
    Toolset,
    ToolsetTag,
)
from holmes.core.tools_utils.token_counting import count_tool_response_tokens
from holmes.core.tools_utils.tool_context_window_limiter import get_pct_token_count
from holmes.plugins.toolsets.consts import STANDARD_END_DATETIME_TOOL_PARAM_DESCRIPTION
from holmes.plugins.toolsets.prometheus.utils import parse_duration_to_seconds
from holmes.plugins.toolsets.service_discovery import PrometheusDiscovery
from holmes.plugins.toolsets.utils import (
    get_param_or_raise,
    process_timestamps_to_rfc3339,
    standard_start_datetime_tool_param_description,
    toolset_name_for_one_liner,
)
from holmes.utils.cache import TTLCache
from holmes.common.env_vars import IS_OPENSHIFT, MAX_GRAPH_POINTS
from holmes.common.openshift import load_openshift_token
from holmes.plugins.toolsets.logging_utils.logging_api import (
    DEFAULT_GRAPH_TIME_SPAN_SECONDS,
)

PROMETHEUS_RULES_CACHE_KEY = "cached_prometheus_rules"
PROMETHEUS_METADATA_API_LIMIT = 100  # Default limit for Prometheus metadata APIs (series, labels, metadata) to prevent overwhelming responses
# Default timeout values for PromQL queries
DEFAULT_QUERY_TIMEOUT_SECONDS = 20
MAX_QUERY_TIMEOUT_SECONDS = 180
# Default timeout for metadata API calls (discovery endpoints)
DEFAULT_METADATA_TIMEOUT_SECONDS = 20
MAX_METADATA_TIMEOUT_SECONDS = 60
# Default time window for metadata APIs (in hours)
DEFAULT_METADATA_TIME_WINDOW_HRS = 1


class PrometheusConfig(BaseModel):
    # URL is optional because it can be set with an env var
    prometheus_url: Optional[str]
    healthcheck: str = "-/healthy"

    # New config for default time window for metadata APIs
    default_metadata_time_window_hrs: int = DEFAULT_METADATA_TIME_WINDOW_HRS  # Default: only show metrics active in the last hour

    # Query timeout configuration
    default_query_timeout_seconds: int = (
        DEFAULT_QUERY_TIMEOUT_SECONDS  # Default timeout for PromQL queries
    )
    max_query_timeout_seconds: int = (
        MAX_QUERY_TIMEOUT_SECONDS  # Maximum allowed timeout for PromQL queries
    )

    # Metadata API timeout configuration
    default_metadata_timeout_seconds: int = (
        DEFAULT_METADATA_TIMEOUT_SECONDS  # Default timeout for metadata/discovery APIs
    )
    max_metadata_timeout_seconds: int = (
        MAX_METADATA_TIMEOUT_SECONDS  # Maximum allowed timeout for metadata APIs
    )

    # DEPRECATED: These config values are deprecated and will be removed in a future version
    # Using None as default so we can detect if user explicitly set them
    metrics_labels_time_window_hrs: Optional[int] = (
        None  # DEPRECATED - use default_metadata_time_window_hrs instead
    )
    metrics_labels_cache_duration_hrs: Optional[int] = (
        None  # DEPRECATED - no longer used
    )
    fetch_labels_with_labels_api: Optional[bool] = None  # DEPRECATED - no longer used
    fetch_metadata_with_series_api: Optional[bool] = None  # DEPRECATED - no longer used

    tool_calls_return_data: bool = True
    headers: Dict = Field(default_factory=dict)
    rules_cache_duration_seconds: Optional[int] = 1800  # 30 minutes
    additional_labels: Optional[Dict[str, str]] = None
    prometheus_ssl_enabled: bool = True

    # Custom limit to the max number of tokens that a query result can take to proactively
    #   prevent token limit issues. Expressed in % of the model's context window.
    # This limit only overrides the global limit for all tools  (TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT)
    #   if it is lower.
    query_response_size_limit_pct: Optional[int] = None

    @field_validator("prometheus_url")
    def ensure_trailing_slash(cls, v: Optional[str]) -> Optional[str]:
        if v is not None and not v.endswith("/"):
            return v + "/"
        return v

    @model_validator(mode="after")
    def validate_prom_config(self):
        # Check for deprecated config values and print warnings
        deprecated_configs = []
        if self.metrics_labels_time_window_hrs is not None:  # Check if explicitly set
            deprecated_configs.append(
                "metrics_labels_time_window_hrs (use default_metadata_time_window_hrs instead)"
            )
        if (
            self.metrics_labels_cache_duration_hrs is not None
        ):  # Check if explicitly set
            deprecated_configs.append("metrics_labels_cache_duration_hrs")
        if self.fetch_labels_with_labels_api is not None:  # Check if explicitly set
            deprecated_configs.append("fetch_labels_with_labels_api")
        if self.fetch_metadata_with_series_api is not None:  # Check if explicitly set
            deprecated_configs.append("fetch_metadata_with_series_api")

        if deprecated_configs:
            logging.warning(
                f"WARNING: The following Prometheus config values are deprecated and will be removed in a future version: "
                f"{', '.join(deprecated_configs)}. These configs no longer affect behavior."
            )
        # If openshift is enabled, and the user didn't configure auth headers, we will try to load the token from the service account.
        if IS_OPENSHIFT:
            if self.healthcheck == "-/healthy":
                self.healthcheck = "api/v1/query?query=up"

            if self.headers.get("Authorization"):
                return self

            openshift_token = load_openshift_token()
            if openshift_token:
                logging.info("Using openshift token for prometheus toolset auth")
                self.headers["Authorization"] = f"Bearer {openshift_token}"

        return self

    def is_amp(self) -> bool:
        return False


class AMPConfig(PrometheusConfig):
    aws_access_key: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_region: str
    aws_service_name: str = "aps"
    healthcheck: str = "api/v1/query?query=up"
    prometheus_ssl_enabled: bool = False
    assume_role_arn: Optional[str] = None

    # Refresh the AWS client (and its STS creds) every N seconds (default: 15 minutes)
    refresh_interval_seconds: int = 900

    _aws_client: Optional[AWSPrometheusConnect] = None
    _aws_client_created_at: float = 0.0

    def is_amp(self) -> bool:
        return True

    def _should_refresh_client(self) -> bool:
        if not self._aws_client:
            return True
        return (
            time.time() - self._aws_client_created_at
        ) >= self.refresh_interval_seconds

    def get_aws_client(self) -> Optional[AWSPrometheusConnect]:
        if not self._aws_client or self._should_refresh_client():
            try:
                base_config = BasePrometheusConfig(
                    url=self.prometheus_url,
                    disable_ssl=not self.prometheus_ssl_enabled,
                    additional_labels=self.additional_labels,
                )
                self._aws_client = AWSPrometheusConnect(
                    access_key=self.aws_access_key,
                    secret_key=self.aws_secret_access_key,
                    token=None,
                    region=self.aws_region,
                    service_name=self.aws_service_name,
                    assume_role_arn=self.assume_role_arn,
                    config=base_config,
                )
                self._aws_client_created_at = time.time()
            except Exception:
                logging.exception("Failed to create/refresh AWS client")
                return self._aws_client
        return self._aws_client


class BasePrometheusTool(Tool):
    toolset: "PrometheusToolset"


def do_request(
    config,  # PrometheusConfig | AMPConfig
    url: str,
    params: Optional[Dict] = None,
    data: Optional[Dict] = None,
    timeout: int = 60,
    verify: Optional[bool] = None,
    headers: Optional[Dict] = None,
    method: str = "GET",
) -> requests.Response:
    """
    Route a request through either:
      - AWSPrometheusConnect (SigV4) when config is AMPConfig
      - plain requests otherwise

    method defaults to GET so callers can omit it for reads.
    """
    if verify is None:
        verify = config.prometheus_ssl_enabled
    if headers is None:
        headers = config.headers or {}

    if isinstance(config, AMPConfig):
        client = config.get_aws_client()  # cached AWSPrometheusConnect
        # Note: timeout parameter is not supported by prometrix's signed_request
        # AWS/AMP requests will not respect the timeout setting
        return client.signed_request(  # type: ignore
            method=method,
            url=url,
            data=data,
            params=params,
            verify=verify,
            headers=headers,
        )

    # Non-AMP: plain HTTP
    return requests.request(
        method=method,
        url=url,
        headers=headers,
        params=params,
        data=data,
        timeout=timeout,
        verify=verify,
    )


def result_has_data(result: Dict) -> bool:
    data = result.get("data", {})
    if len(data.get("result", [])) > 0:
        return True
    return False


def adjust_step_for_max_points(
    start_timestamp: str,
    end_timestamp: str,
    step: Optional[float] = None,
    max_points_override: Optional[float] = None,
) -> float:
    """
    Adjusts the step parameter to ensure the number of data points doesn't exceed max_points.

    Args:
        start_timestamp: RFC3339 formatted start time
        end_timestamp: RFC3339 formatted end time
        step: The requested step duration in seconds (None for auto-calculation)
        max_points_override: Optional override for max points (must be <= MAX_GRAPH_POINTS)

    Returns:
        Adjusted step value in seconds that ensures points <= max_points
    """
    # Use override if provided and valid, otherwise use default
    max_points = MAX_GRAPH_POINTS
    if max_points_override is not None:
        if max_points_override > MAX_GRAPH_POINTS:
            logging.warning(
                f"max_points override ({max_points_override}) exceeds system limit ({MAX_GRAPH_POINTS}), using {MAX_GRAPH_POINTS}"
            )
            max_points = MAX_GRAPH_POINTS
        elif max_points_override < 1:
            logging.warning(
                f"max_points override ({max_points_override}) is invalid, using default {MAX_GRAPH_POINTS}"
            )
            max_points = MAX_GRAPH_POINTS
        else:
            max_points = max_points_override
            logging.debug(f"Using max_points override: {max_points}")

    start_dt = dateutil.parser.parse(start_timestamp)
    end_dt = dateutil.parser.parse(end_timestamp)

    time_range_seconds = (end_dt - start_dt).total_seconds()

    # If no step provided, calculate a reasonable default
    # Aim for ~60 data points across the time range (1 per minute for hourly, etc)
    if step is None:
        step = max(1, time_range_seconds / 60)
        logging.debug(
            f"No step provided, defaulting to {step}s for {time_range_seconds}s range"
        )

    current_points = time_range_seconds / step

    # If current points exceed max, adjust the step
    if current_points > max_points:
        adjusted_step = time_range_seconds / max_points
        logging.info(
            f"Adjusting step from {step}s to {adjusted_step}s to limit points from {current_points:.0f} to {max_points}"
        )
        return adjusted_step

    return step


def add_prometheus_auth(prometheus_auth_header: Optional[str]) -> Dict[str, Any]:
    results = {}
    if prometheus_auth_header:
        results["Authorization"] = prometheus_auth_header
    return results


def create_data_summary_for_large_result(
    result_data: Dict, query: str, data_size_tokens: int, is_range_query: bool = False
) -> Dict[str, Any]:
    """
    Create a summary for large Prometheus results instead of returning full data.

    Args:
        result_data: The Prometheus data result
        query: The original PromQL query
        data_size_tokens: Size of the data in tokens
        is_range_query: Whether this is a range query (vs instant query)

    Returns:
        Dictionary with summary information and suggestions
    """
    if is_range_query:
        series_list = result_data.get("result", [])
        num_items = len(series_list)

        # Calculate exact total data points across all series
        total_points = 0
        for series in series_list:  # Iterate through ALL series for exact count
            points = len(series.get("values", []))
            total_points += points

        # Analyze label keys and their cardinality
        label_cardinality: Dict[str, set] = {}
        for series in series_list:
            metric = series.get("metric", {})
            for label_key, label_value in metric.items():
                if label_key not in label_cardinality:
                    label_cardinality[label_key] = set()
                label_cardinality[label_key].add(label_value)

        # Convert sets to counts for the summary
        label_summary = {
            label: len(values) for label, values in label_cardinality.items()
        }
        # Sort by cardinality (highest first) for better insights
        label_summary = dict(
            sorted(label_summary.items(), key=lambda x: x[1], reverse=True)
        )

        return {
            "message": f"Data too large to return ({data_size_tokens:,} tokens). Query returned {num_items} time series with {total_points:,} total data points.",
            "series_count": num_items,
            "total_data_points": total_points,
            "data_size_tokens": data_size_tokens,
            "label_cardinality": label_summary,
            "suggestion": f'Consider using topk({min(5, num_items)}, {query}) to limit results to the top {min(5, num_items)} series. To also capture remaining data as \'other\': topk({min(5, num_items)}, {query}) or label_replace((sum({query}) - sum(topk({min(5, num_items)}, {query}))), "pod", "other", "", "")',
        }
    else:
        # Instant query
        result_type = result_data.get("resultType", "")
        result_list = result_data.get("result", [])
        num_items = len(result_list)

        # Analyze label keys and their cardinality
        instant_label_cardinality: Dict[str, set] = {}
        for item in result_list:
            if isinstance(item, dict):
                metric = item.get("metric", {})
                for label_key, label_value in metric.items():
                    if label_key not in instant_label_cardinality:
                        instant_label_cardinality[label_key] = set()
                    instant_label_cardinality[label_key].add(label_value)

        # Convert sets to counts for the summary
        label_summary = {
            label: len(values) for label, values in instant_label_cardinality.items()
        }
        # Sort by cardinality (highest first) for better insights
        label_summary = dict(
            sorted(label_summary.items(), key=lambda x: x[1], reverse=True)
        )

        return {
            "message": f"Data too large to return ({data_size_tokens:,} tokens). Query returned {num_items} results.",
            "result_count": num_items,
            "result_type": result_type,
            "data_size_tokens": data_size_tokens,
            "label_cardinality": label_summary,
            "suggestion": f'Consider using topk({min(5, num_items)}, {query}) to limit results. To also capture remaining data as \'other\': topk({min(5, num_items)}, {query}) or label_replace((sum({query}) - sum(topk({min(5, num_items)}, {query}))), "instance", "other", "", "")',
        }


class MetricsBasedResponse(BaseModel):
    status: str
    error_message: Optional[str] = None
    data: Optional[str] = None
    tool_name: str
    description: str
    query: str
    start: Optional[str] = None
    end: Optional[str] = None
    step: Optional[float] = None
    output_type: Optional[str] = None
    data_summary: Optional[dict[str, Any]] = None


def create_structured_tool_result(
    params: dict, response: MetricsBasedResponse
) -> StructuredToolResult:
    status = StructuredToolResultStatus.SUCCESS
    error = None
    if response.error_message or response.status.lower() in ("failed", "error"):
        status = StructuredToolResultStatus.ERROR
        error = (
            response.error_message
            if response.error_message
            else "Unknown Prometheus error"
        )
    elif not response.data:
        status = StructuredToolResultStatus.NO_DATA

    return StructuredToolResult(
        status=status,
        data=response.model_dump_json(indent=2),
        params=params,
        error=error,
    )


class ListPrometheusRules(BasePrometheusTool):
    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="list_prometheus_rules",
            description="List all defined Prometheus rules (api/v1/rules). Will show the Prometheus rules description, expression and annotations",
            parameters={},
            toolset=toolset,
        )
        self._cache = None

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        if self.toolset.config.is_amp():
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Tool not supported in AMP",
                params=params,
            )
        if not self._cache and self.toolset.config.rules_cache_duration_seconds:
            self._cache = TTLCache(self.toolset.config.rules_cache_duration_seconds)  # type: ignore
        try:
            if self._cache:
                cached_rules = self._cache.get(PROMETHEUS_RULES_CACHE_KEY)
                if cached_rules:
                    logging.debug("rules returned from cache")

                    return StructuredToolResult(
                        status=StructuredToolResultStatus.SUCCESS,
                        data=cached_rules,
                        params=params,
                    )

            prometheus_url = self.toolset.config.prometheus_url

            rules_url = urljoin(prometheus_url, "api/v1/rules")

            rules_response = do_request(
                config=self.toolset.config,
                url=rules_url,
                params=params,
                timeout=40,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            rules_response.raise_for_status()
            data = rules_response.json()["data"]

            if self._cache:
                self._cache.set(PROMETHEUS_RULES_CACHE_KEY, data)
            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except requests.Timeout:
            logging.warning("Timeout while fetching prometheus rules", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Request timed out while fetching rules",
                params=params,
            )
        except RequestException as e:
            logging.warning("Failed to fetch prometheus rules", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Network error while fetching rules: {str(e)}",
                params=params,
            )
        except Exception as e:
            logging.warning("Failed to process prometheus rules", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Unexpected error: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Fetch Rules"


class GetMetricNames(BasePrometheusTool):
    """Thin wrapper around /api/v1/label/__name__/values - the fastest way to discover metric names"""

    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="get_metric_names",
            description=(
                "Get list of metric names using /api/v1/label/__name__/values. "
                "FASTEST method for metric discovery when you need to explore available metrics. "
                f"Returns up to {PROMETHEUS_METADATA_API_LIMIT} unique metric names (limit={PROMETHEUS_METADATA_API_LIMIT}). If {PROMETHEUS_METADATA_API_LIMIT} results returned, more may exist - use a more specific filter. "
                f"ALWAYS use match[] parameter to filter metrics - without it you'll get random {PROMETHEUS_METADATA_API_LIMIT} metrics which is rarely useful. "
                "Note: Does not return metric metadata (type, description, labels). "
                "By default returns metrics active in the last 1 hour (configurable via default_metadata_time_window_hrs)."
            ),
            parameters={
                "match": ToolParameter(
                    description=(
                        "REQUIRED: PromQL selector to filter metrics. Use regex OR (|) to check multiple patterns in one call - much faster than multiple calls! Examples: "
                        "'{__name__=~\"node_cpu.*|node_memory.*|node_disk.*\"}' for all node resource metrics, "
                        "'{__name__=~\"container_cpu.*|container_memory.*|container_network.*\"}' for all container metrics, "
                        "'{__name__=~\"kube_pod.*|kube_deployment.*|kube_service.*\"}' for multiple Kubernetes object metrics, "
                        "'{__name__=~\".*cpu.*|.*memory.*|.*disk.*\"}' for all resource metrics, "
                        "'{namespace=~\"kube-system|default|monitoring\"}' for metrics from multiple namespaces, "
                        "'{job=~\"prometheus|node-exporter|kube-state-metrics\"}' for metrics from multiple jobs."
                    ),
                    type="string",
                    required=True,
                ),
                "start": ToolParameter(
                    description="Start timestamp (RFC3339 or Unix). Default: 1 hour ago",
                    type="string",
                    required=False,
                ),
                "end": ToolParameter(
                    description="End timestamp (RFC3339 or Unix). Default: now",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            match_param = params.get("match")
            if not match_param:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error="Match parameter is required to filter metrics",
                    params=params,
                )

            url = urljoin(
                self.toolset.config.prometheus_url, "api/v1/label/__name__/values"
            )
            query_params = {
                "limit": str(PROMETHEUS_METADATA_API_LIMIT),
                "match[]": match_param,
            }

            # Add time parameters - use provided values or defaults
            if params.get("end"):
                query_params["end"] = params["end"]
            else:
                query_params["end"] = str(int(time.time()))

            if params.get("start"):
                query_params["start"] = params["start"]
            elif self.toolset.config.default_metadata_time_window_hrs:
                # Use default time window
                query_params["start"] = str(
                    int(time.time())
                    - (self.toolset.config.default_metadata_time_window_hrs * 3600)
                )

            response = do_request(
                config=self.toolset.config,
                url=url,
                params=query_params,
                timeout=self.toolset.config.default_metadata_timeout_seconds,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            response.raise_for_status()
            data = response.json()

            # Check if results were truncated
            if (
                "data" in data
                and isinstance(data["data"], list)
                and len(data["data"]) == PROMETHEUS_METADATA_API_LIMIT
            ):
                data["_truncated"] = True
                data["_message"] = (
                    f"Results truncated at limit={PROMETHEUS_METADATA_API_LIMIT}. Use a more specific match filter to see additional metrics."
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except Exception as e:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=str(e),
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Get Metric Names"


class GetLabelValues(BasePrometheusTool):
    """Get values for a specific label across all metrics"""

    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="get_label_values",
            description=(
                "Get all values for a specific label using /api/v1/label/{label}/values. "
                "Use this to discover pods, namespaces, jobs, instances, etc. "
                f"Returns up to {PROMETHEUS_METADATA_API_LIMIT} unique values (limit={PROMETHEUS_METADATA_API_LIMIT}). If {PROMETHEUS_METADATA_API_LIMIT} results returned, more may exist - use match[] to filter. "
                "Supports optional match[] parameter to filter. "
                "By default returns values from metrics active in the last 1 hour (configurable via default_metadata_time_window_hrs)."
            ),
            parameters={
                "label": ToolParameter(
                    description="Label name to get values for (e.g., 'pod', 'namespace', 'job', 'instance')",
                    type="string",
                    required=True,
                ),
                "match": ToolParameter(
                    description=(
                        "Optional PromQL selector to filter (e.g., '{__name__=~\"kube.*\"}', "
                        "'{namespace=\"default\"}')."
                    ),
                    type="string",
                    required=False,
                ),
                "start": ToolParameter(
                    description="Start timestamp (RFC3339 or Unix). Default: 1 hour ago",
                    type="string",
                    required=False,
                ),
                "end": ToolParameter(
                    description="End timestamp (RFC3339 or Unix). Default: now",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            label = params.get("label")
            if not label:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error="Label parameter is required",
                    params=params,
                )

            url = urljoin(
                self.toolset.config.prometheus_url, f"api/v1/label/{label}/values"
            )
            query_params = {"limit": str(PROMETHEUS_METADATA_API_LIMIT)}
            if params.get("match"):
                query_params["match[]"] = params["match"]

            # Add time parameters - use provided values or defaults
            if params.get("end"):
                query_params["end"] = params["end"]
            else:
                query_params["end"] = str(int(time.time()))

            if params.get("start"):
                query_params["start"] = params["start"]
            elif self.toolset.config.default_metadata_time_window_hrs:
                # Use default time window
                query_params["start"] = str(
                    int(time.time())
                    - (self.toolset.config.default_metadata_time_window_hrs * 3600)
                )

            response = do_request(
                config=self.toolset.config,
                url=url,
                params=query_params,
                timeout=self.toolset.config.default_metadata_timeout_seconds,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            response.raise_for_status()
            data = response.json()

            # Check if results were truncated
            if (
                "data" in data
                and isinstance(data["data"], list)
                and len(data["data"]) == PROMETHEUS_METADATA_API_LIMIT
            ):
                data["_truncated"] = True
                data["_message"] = (
                    f"Results truncated at limit={PROMETHEUS_METADATA_API_LIMIT}. Use match[] parameter to filter label '{label}' values."
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except Exception as e:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=str(e),
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        label = params.get("label", "")
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Get {label} Values"


class GetAllLabels(BasePrometheusTool):
    """Get all label names that exist in Prometheus"""

    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="get_all_labels",
            description=(
                "Get list of all label names using /api/v1/labels. "
                "Use this to discover what labels are available across all metrics. "
                f"Returns up to {PROMETHEUS_METADATA_API_LIMIT} label names (limit={PROMETHEUS_METADATA_API_LIMIT}). If {PROMETHEUS_METADATA_API_LIMIT} results returned, more may exist - use match[] to filter. "
                "Supports optional match[] parameter to filter. "
                "By default returns labels from metrics active in the last 1 hour (configurable via default_metadata_time_window_hrs)."
            ),
            parameters={
                "match": ToolParameter(
                    description=(
                        "Optional PromQL selector to filter (e.g., '{__name__=~\"kube.*\"}', "
                        "'{job=\"prometheus\"}')."
                    ),
                    type="string",
                    required=False,
                ),
                "start": ToolParameter(
                    description="Start timestamp (RFC3339 or Unix). Default: 1 hour ago",
                    type="string",
                    required=False,
                ),
                "end": ToolParameter(
                    description="End timestamp (RFC3339 or Unix). Default: now",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            url = urljoin(self.toolset.config.prometheus_url, "api/v1/labels")
            query_params = {"limit": str(PROMETHEUS_METADATA_API_LIMIT)}
            if params.get("match"):
                query_params["match[]"] = params["match"]

            # Add time parameters - use provided values or defaults
            if params.get("end"):
                query_params["end"] = params["end"]
            else:
                query_params["end"] = str(int(time.time()))

            if params.get("start"):
                query_params["start"] = params["start"]
            elif self.toolset.config.default_metadata_time_window_hrs:
                # Use default time window
                query_params["start"] = str(
                    int(time.time())
                    - (self.toolset.config.default_metadata_time_window_hrs * 3600)
                )

            response = do_request(
                config=self.toolset.config,
                url=url,
                params=query_params,
                timeout=self.toolset.config.default_metadata_timeout_seconds,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            response.raise_for_status()
            data = response.json()

            # Check if results were truncated
            if (
                "data" in data
                and isinstance(data["data"], list)
                and len(data["data"]) == PROMETHEUS_METADATA_API_LIMIT
            ):
                data["_truncated"] = True
                data["_message"] = (
                    f"Results truncated at limit={PROMETHEUS_METADATA_API_LIMIT}. Use match[] parameter to filter labels."
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except Exception as e:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=str(e),
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Get All Labels"


class GetSeries(BasePrometheusTool):
    """Get time series matching a selector"""

    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="get_series",
            description=(
                "Get time series using /api/v1/series. "
                "Returns label sets for all time series matching the selector. "
                "SLOWER than other discovery methods - use only when you need full label sets. "
                f"Returns up to {PROMETHEUS_METADATA_API_LIMIT} series (limit={PROMETHEUS_METADATA_API_LIMIT}). If {PROMETHEUS_METADATA_API_LIMIT} results returned, more series exist - use more specific selector. "
                "Requires match[] parameter with PromQL selector. "
                "By default returns series active in the last 1 hour (configurable via default_metadata_time_window_hrs)."
            ),
            parameters={
                "match": ToolParameter(
                    description=(
                        "PromQL selector to match series (e.g., 'up', 'node_cpu_seconds_total', "
                        "'{__name__=~\"node.*\"}', '{job=\"prometheus\"}', "
                        '\'{__name__="up",job="prometheus"}\').'
                    ),
                    type="string",
                    required=True,
                ),
                "start": ToolParameter(
                    description="Start timestamp (RFC3339 or Unix). Default: 1 hour ago",
                    type="string",
                    required=False,
                ),
                "end": ToolParameter(
                    description="End timestamp (RFC3339 or Unix). Default: now",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            match = params.get("match")
            if not match:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error="Match parameter is required",
                    params=params,
                )

            url = urljoin(self.toolset.config.prometheus_url, "api/v1/series")
            query_params = {
                "match[]": match,
                "limit": str(PROMETHEUS_METADATA_API_LIMIT),
            }

            # Add time parameters - use provided values or defaults
            if params.get("end"):
                query_params["end"] = params["end"]
            else:
                query_params["end"] = str(int(time.time()))

            if params.get("start"):
                query_params["start"] = params["start"]
            elif self.toolset.config.default_metadata_time_window_hrs:
                # Use default time window
                query_params["start"] = str(
                    int(time.time())
                    - (self.toolset.config.default_metadata_time_window_hrs * 3600)
                )

            response = do_request(
                config=self.toolset.config,
                url=url,
                params=query_params,
                timeout=self.toolset.config.default_metadata_timeout_seconds,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            response.raise_for_status()
            data = response.json()

            # Check if results were truncated
            if (
                "data" in data
                and isinstance(data["data"], list)
                and len(data["data"]) == PROMETHEUS_METADATA_API_LIMIT
            ):
                data["_truncated"] = True
                data["_message"] = (
                    f"Results truncated at limit={PROMETHEUS_METADATA_API_LIMIT}. Use a more specific match selector to see additional series."
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except Exception as e:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=str(e),
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Get Series"


class GetMetricMetadata(BasePrometheusTool):
    """Get metadata (type, description, unit) for metrics"""

    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="get_metric_metadata",
            description=(
                "Get metric metadata using /api/v1/metadata. "
                "Returns type, help text, and unit for metrics. "
                "Use after discovering metric names to get their descriptions. "
                f"Returns up to {PROMETHEUS_METADATA_API_LIMIT} metrics (limit={PROMETHEUS_METADATA_API_LIMIT}). If {PROMETHEUS_METADATA_API_LIMIT} results returned, more may exist - filter by specific metric name. "
                "Supports optional metric name filter."
            ),
            parameters={
                "metric": ToolParameter(
                    description=(
                        "Optional metric name to filter (e.g., 'up', 'node_cpu_seconds_total'). "
                        "If not provided, returns metadata for all metrics."
                    ),
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            url = urljoin(self.toolset.config.prometheus_url, "api/v1/metadata")
            query_params = {"limit": str(PROMETHEUS_METADATA_API_LIMIT)}

            if params.get("metric"):
                query_params["metric"] = params["metric"]

            response = do_request(
                config=self.toolset.config,
                url=url,
                params=query_params,
                timeout=self.toolset.config.default_metadata_timeout_seconds,
                verify=self.toolset.config.prometheus_ssl_enabled,
                headers=self.toolset.config.headers,
                method="GET",
            )
            response.raise_for_status()
            data = response.json()

            # Check if results were truncated (metadata endpoint returns a dict, not a list)
            if (
                "data" in data
                and isinstance(data["data"], dict)
                and len(data["data"]) == PROMETHEUS_METADATA_API_LIMIT
            ):
                data["_truncated"] = True
                data["_message"] = (
                    f"Results truncated at limit={PROMETHEUS_METADATA_API_LIMIT}. Use metric parameter to filter by specific metric name."
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )
        except Exception as e:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=str(e),
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        metric = params.get("metric", "all")
        return (
            f"{toolset_name_for_one_liner(self.toolset.name)}: Get Metadata ({metric})"
        )


class ExecuteInstantQuery(BasePrometheusTool):
    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="execute_prometheus_instant_query",
            description=(
                f"Execute an instant PromQL query (single point in time). "
                f"Default timeout is {DEFAULT_QUERY_TIMEOUT_SECONDS} seconds "
                f"but can be increased up to {MAX_QUERY_TIMEOUT_SECONDS} seconds for complex/slow queries."
            ),
            parameters={
                "query": ToolParameter(
                    description="The PromQL query",
                    type="string",
                    required=True,
                ),
                "description": ToolParameter(
                    description="Describes the query",
                    type="string",
                    required=True,
                ),
                "timeout": ToolParameter(
                    description=(
                        f"Query timeout in seconds. Default: {DEFAULT_QUERY_TIMEOUT_SECONDS}. "
                        f"Maximum: {MAX_QUERY_TIMEOUT_SECONDS}. "
                        f"Increase for complex queries that may take longer."
                    ),
                    type="number",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )
        try:
            query = params.get("query", "")
            description = params.get("description", "")

            url = urljoin(self.toolset.config.prometheus_url, "api/v1/query")

            payload = {"query": query}

            # Get timeout parameter and enforce limits
            default_timeout = self.toolset.config.default_query_timeout_seconds
            max_timeout = self.toolset.config.max_query_timeout_seconds
            timeout = params.get("timeout", default_timeout)
            if timeout > max_timeout:
                timeout = max_timeout
                logging.warning(
                    f"Timeout requested ({params.get('timeout')}) exceeds maximum ({max_timeout}s), using {max_timeout}s"
                )
            elif timeout < 1:
                timeout = default_timeout  # Min 1 second, but use default if invalid

            response = do_request(
                config=self.toolset.config,
                url=url,
                headers=self.toolset.config.headers,
                data=payload,
                timeout=timeout,
                verify=self.toolset.config.prometheus_ssl_enabled,
                method="POST",
            )

            if response.status_code == 200:
                data = response.json()
                status = data.get("status")
                error_message = None
                if status == "success" and not result_has_data(data):
                    status = "Failed"
                    error_message = (
                        "The prometheus query returned no result. Is the query correct?"
                    )
                response_data = MetricsBasedResponse(
                    status=status,
                    error_message=error_message,
                    tool_name=self.name,
                    description=description,
                    query=query,
                )
                structured_tool_result: StructuredToolResult
                # Check if data should be included based on size
                if self.toolset.config.tool_calls_return_data:
                    result_data = data.get("data", {})
                    response_data.data = result_data

                    structured_tool_result = create_structured_tool_result(
                        params=params, response=response_data
                    )
                    tool_call_id = context.tool_call_id
                    tool_name = context.tool_name
                    token_count = count_tool_response_tokens(
                        llm=context.llm,
                        structured_tool_result=structured_tool_result,
                        tool_call_id=tool_call_id,
                        tool_name=tool_name,
                    )

                    token_limit = context.max_token_count
                    if self.toolset.config.query_response_size_limit_pct:
                        custom_token_limit = get_pct_token_count(
                            percent_of_total_context_window=self.toolset.config.query_response_size_limit_pct,
                            llm=context.llm,
                        )
                        if custom_token_limit < token_limit:
                            token_limit = custom_token_limit

                    # Provide summary if data is too large
                    if token_count > token_limit:
                        response_data.data = None
                        response_data.data_summary = (
                            create_data_summary_for_large_result(
                                result_data,
                                query,
                                token_count,
                                is_range_query=False,
                            )
                        )
                        logging.info(
                            f"Prometheus instant query returned large dataset: "
                            f"{response_data.data_summary.get('result_count', 0)} results, "
                            f"{token_count:,} tokens (limit: {token_limit:,}). "
                            f"Returning summary instead of full data."
                        )
                        # Also add token info to the summary for debugging
                        response_data.data_summary["_debug_info"] = (
                            f"Data size: {token_count:,} tokens exceeded limit of {token_limit:,} tokens"
                        )
                    else:
                        response_data.data = result_data

                structured_tool_result = create_structured_tool_result(
                    params=params, response=response_data
                )
                return structured_tool_result

            # Handle known Prometheus error status codes
            error_msg = "Unknown error occurred"
            if response.status_code in [400, 429]:
                try:
                    error_data = response.json()
                    error_msg = error_data.get(
                        "error", error_data.get("message", str(response.content))
                    )
                except json.JSONDecodeError:
                    pass
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error=f"Query execution failed. HTTP {response.status_code}: {error_msg}",
                    params=params,
                )

            # For other status codes, just return the status code and content
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Query execution failed with unexpected status code: {response.status_code}. Response: {str(response.content)}",
                params=params,
            )

        except RequestException as e:
            logging.info("Failed to connect to Prometheus", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Connection error to Prometheus: {str(e)}",
                params=params,
            )
        except Exception as e:
            logging.info("Failed to connect to Prometheus", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Unexpected error executing query: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        description = params.get("description", "")
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Query ({description})"


class ExecuteRangeQuery(BasePrometheusTool):
    def __init__(self, toolset: "PrometheusToolset"):
        super().__init__(
            name="execute_prometheus_range_query",
            description=(
                f"Generates a graph and Execute a PromQL range query. "
                f"Default timeout is {DEFAULT_QUERY_TIMEOUT_SECONDS} seconds "
                f"but can be increased up to {MAX_QUERY_TIMEOUT_SECONDS} seconds for complex/slow queries. "
                f"Default time range is last 1 hour."
            ),
            parameters={
                "query": ToolParameter(
                    description="The PromQL query",
                    type="string",
                    required=True,
                ),
                "description": ToolParameter(
                    description="Describes the query",
                    type="string",
                    required=True,
                ),
                "start": ToolParameter(
                    description=standard_start_datetime_tool_param_description(
                        DEFAULT_GRAPH_TIME_SPAN_SECONDS
                    ),
                    type="string",
                    required=False,
                ),
                "end": ToolParameter(
                    description=STANDARD_END_DATETIME_TOOL_PARAM_DESCRIPTION,
                    type="string",
                    required=False,
                ),
                "step": ToolParameter(
                    description="Query resolution step width in duration format or float number of seconds",
                    type="number",
                    required=False,
                ),
                "output_type": ToolParameter(
                    description="Specifies how to interpret the Prometheus result. Use 'Plain' for raw values, 'Bytes' to format byte values, 'Percentage' to scale 0–1 values into 0–100%, or 'CPUUsage' to convert values to cores (e.g., 500 becomes 500m, 2000 becomes 2).",
                    type="string",
                    required=True,
                ),
                "timeout": ToolParameter(
                    description=(
                        f"Query timeout in seconds. Default: {DEFAULT_QUERY_TIMEOUT_SECONDS}. "
                        f"Maximum: {MAX_QUERY_TIMEOUT_SECONDS}. "
                        f"Increase for complex queries that may take longer."
                    ),
                    type="number",
                    required=False,
                ),
                "max_points": ToolParameter(
                    description=(
                        f"Maximum number of data points to return. Default: {int(MAX_GRAPH_POINTS)}. "
                        f"Can be reduced to get fewer data points (e.g., 50 for simpler graphs). "
                        f"Cannot exceed system limit of {int(MAX_GRAPH_POINTS)}. "
                        f"If your query would return more points than this limit, the step will be automatically adjusted."
                    ),
                    type="number",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.config or not self.toolset.config.prometheus_url:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error="Prometheus is not configured. Prometheus URL is missing",
                params=params,
            )

        try:
            url = urljoin(self.toolset.config.prometheus_url, "api/v1/query_range")

            query = get_param_or_raise(params, "query")
            (start, end) = process_timestamps_to_rfc3339(
                start_timestamp=params.get("start"),
                end_timestamp=params.get("end"),
                default_time_span_seconds=DEFAULT_GRAPH_TIME_SPAN_SECONDS,
            )
            step = parse_duration_to_seconds(params.get("step"))
            max_points = params.get(
                "max_points"
            )  # Get the optional max_points parameter

            # adjust_step_for_max_points handles None case and converts to float
            step = adjust_step_for_max_points(
                start_timestamp=start,
                end_timestamp=end,
                step=step,
                max_points_override=max_points,
            )

            description = params.get("description", "")
            output_type = params.get("output_type", "Plain")
            payload = {
                "query": query,
                "start": start,
                "end": end,
                "step": step,
            }

            # Get timeout parameter and enforce limits
            default_timeout = self.toolset.config.default_query_timeout_seconds
            max_timeout = self.toolset.config.max_query_timeout_seconds
            timeout = params.get("timeout", default_timeout)
            if timeout > max_timeout:
                timeout = max_timeout
                logging.warning(
                    f"Timeout requested ({params.get('timeout')}) exceeds maximum ({max_timeout}s), using {max_timeout}s"
                )
            elif timeout < 1:
                timeout = default_timeout  # Min 1 second, but use default if invalid

            response = do_request(
                config=self.toolset.config,
                url=url,
                headers=self.toolset.config.headers,
                data=payload,
                timeout=timeout,
                verify=self.toolset.config.prometheus_ssl_enabled,
                method="POST",
            )

            if response.status_code == 200:
                data = response.json()
                status = data.get("status")
                error_message = None
                if status == "success" and not result_has_data(data):
                    status = "Failed"
                    error_message = (
                        "The prometheus query returned no result. Is the query correct?"
                    )
                response_data = MetricsBasedResponse(
                    status=status,
                    error_message=error_message,
                    tool_name=self.name,
                    description=description,
                    query=query,
                    start=start,
                    end=end,
                    step=step,
                    output_type=output_type,
                )

                structured_tool_result: StructuredToolResult

                # Check if data should be included based on size
                if self.toolset.config.tool_calls_return_data:
                    result_data = data.get("data", {})
                    response_data.data = result_data
                    structured_tool_result = create_structured_tool_result(
                        params=params, response=response_data
                    )

                    tool_call_id = context.tool_call_id
                    tool_name = context.tool_name
                    token_count = count_tool_response_tokens(
                        llm=context.llm,
                        structured_tool_result=structured_tool_result,
                        tool_call_id=tool_call_id,
                        tool_name=tool_name,
                    )

                    token_limit = context.max_token_count
                    if self.toolset.config.query_response_size_limit_pct:
                        custom_token_limit = get_pct_token_count(
                            percent_of_total_context_window=self.toolset.config.query_response_size_limit_pct,
                            llm=context.llm,
                        )
                        if custom_token_limit < token_limit:
                            token_limit = custom_token_limit

                    # Provide summary if data is too large
                    if token_count > token_limit:
                        response_data.data = None
                        response_data.data_summary = (
                            create_data_summary_for_large_result(
                                result_data, query, token_count, is_range_query=True
                            )
                        )
                        logging.info(
                            f"Prometheus range query returned large dataset: "
                            f"{response_data.data_summary.get('series_count', 0)} series, "
                            f"{token_count:,} tokens (limit: {token_limit:,}). "
                            f"Returning summary instead of full data."
                        )
                        # Also add character info to the summary for debugging
                        response_data.data_summary["_debug_info"] = (
                            f"Data size: {token_count:,} tokens exceeded limit of {token_limit:,} tokens"
                        )
                    else:
                        response_data.data = result_data

                structured_tool_result = create_structured_tool_result(
                    params=params, response=response_data
                )

                return structured_tool_result

            error_msg = "Unknown error occurred"
            if response.status_code in [400, 429]:
                try:
                    error_data = response.json()
                    error_msg = error_data.get(
                        "error", error_data.get("message", str(response.content))
                    )
                except json.JSONDecodeError:
                    pass
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error=f"Query execution failed. HTTP {response.status_code}: {error_msg}",
                    params=params,
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Query execution failed with unexpected status code: {response.status_code}. Response: {str(response.content)}",
                params=params,
            )

        except RequestException as e:
            logging.info("Failed to connect to Prometheus", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Connection error to Prometheus: {str(e)}",
                params=params,
            )
        except Exception as e:
            logging.info("Failed to connect to Prometheus", exc_info=True)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Unexpected error executing query: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        description = params.get("description", "")
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Query ({description})"


class PrometheusToolset(Toolset):
    config: Optional[Union[PrometheusConfig, AMPConfig]] = None

    def __init__(self):
        super().__init__(
            name="prometheus/metrics",
            description="Prometheus integration to fetch metadata and execute PromQL queries",
            docs_url="https://holmesgpt.dev/data-sources/builtin-toolsets/prometheus/",
            icon_url="https://upload.wikimedia.org/wikipedia/commons/3/38/Prometheus_software_logo.svg",
            prerequisites=[CallablePrerequisite(callable=self.prerequisites_callable)],
            tools=[
                ListPrometheusRules(toolset=self),
                GetMetricNames(toolset=self),
                GetLabelValues(toolset=self),
                GetAllLabels(toolset=self),
                GetSeries(toolset=self),
                GetMetricMetadata(toolset=self),
                ExecuteInstantQuery(toolset=self),
                ExecuteRangeQuery(toolset=self),
            ],
            tags=[
                ToolsetTag.CORE,
            ],
        )
        self._reload_llm_instructions()

    def _reload_llm_instructions(self):
        template_file_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "prometheus_instructions.jinja2")
        )
        self._load_llm_instructions(jinja_template=f"file://{template_file_path}")

    def determine_prometheus_class(
        self, config: dict[str, Any]
    ) -> Type[Union[PrometheusConfig, AMPConfig]]:
        has_aws_fields = "aws_region" in config
        return AMPConfig if has_aws_fields else PrometheusConfig

    def prerequisites_callable(self, config: dict[str, Any]) -> Tuple[bool, str]:
        try:
            if config:
                config_cls = self.determine_prometheus_class(config)
                self.config = config_cls(**config)  # type: ignore

                self._reload_llm_instructions()
                return self._is_healthy()
        except Exception:
            logging.exception("Failed to create prometheus config")
            return False, "Failed to create prometheus config"
        try:
            prometheus_url = os.environ.get("PROMETHEUS_URL")
            if not prometheus_url:
                prometheus_url = self.auto_detect_prometheus_url()
                if not prometheus_url:
                    return (
                        False,
                        "Unable to auto-detect prometheus. Define prometheus_url in the configuration for tool prometheus/metrics",
                    )

            self.config = PrometheusConfig(
                prometheus_url=prometheus_url,
                headers=add_prometheus_auth(os.environ.get("PROMETHEUS_AUTH_HEADER")),
            )
            logging.info(f"Prometheus auto discovered at url {prometheus_url}")
            self._reload_llm_instructions()
            return self._is_healthy()
        except Exception as e:
            logging.exception("Failed to set up prometheus")
            return False, str(e)

    def auto_detect_prometheus_url(self) -> Optional[str]:
        url: Optional[str] = PrometheusDiscovery.find_prometheus_url()
        if not url:
            url = PrometheusDiscovery.find_vm_url()

        return url

    def _is_healthy(self) -> Tuple[bool, str]:
        if (
            not hasattr(self, "config")
            or not self.config
            or not self.config.prometheus_url
        ):
            return (
                False,
                f"Toolset {self.name} failed to initialize because prometheus is not configured correctly",
            )

        url = urljoin(self.config.prometheus_url, self.config.healthcheck)
        try:
            response = do_request(
                config=self.config,
                url=url,
                headers=self.config.headers,
                timeout=10,
                verify=self.config.prometheus_ssl_enabled,
                method="GET",
            )

            if response.status_code == 200:
                return True, ""
            else:
                return (
                    False,
                    f"Failed to connect to Prometheus at {url}: HTTP {response.status_code}",
                )

        except Exception as e:
            logging.debug("Failed to initialize Prometheus", exc_info=True)
            return (
                False,
                f"Failed to initialize using url={url}. Unexpected error: {str(e)}",
            )

    def get_example_config(self):
        example_config = PrometheusConfig(
            prometheus_url="http://robusta-kube-prometheus-st-prometheus:9090"
        )
        return example_config.model_dump()



import logging
import requests  # type: ignore
from typing import Tuple
import backoff

from holmes.plugins.toolsets.grafana.common import (
    GrafanaConfig,
    build_headers,
)


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_tries=2,
    giveup=lambda e: isinstance(e, requests.exceptions.HTTPError)
    and e.response.status_code < 500,
)
def _try_health_url(url: str, headers: dict) -> None:
    response = requests.get(url, headers=headers, timeout=5)
    response.raise_for_status()


def grafana_health_check(config: GrafanaConfig) -> Tuple[bool, str]:
    """
    Tests a healthcheck url for grafna loki.
    1. When using grafana as proxy, grafana_datasource_uid is provided, use the data source health url (docs are added).
    2. When using loki directly there are two cases.
        a. Using loki cloud, health check is provided on the base url.
        b. Using local loki, uses url/healthcheck default is url/ready
        c. This function tries both direct loki cases for the user.
    """
    health_urls = []
    if config.grafana_datasource_uid:
        # https://grafana.com/docs/grafana/latest/developers/http_api/data_source/#check-data-source-health
        health_urls.append(
            f"{config.url}/api/datasources/uid/{config.grafana_datasource_uid}/health"
        )
    else:
        health_urls.append(f"{config.url}/{config.healthcheck}")
        health_urls.append(config.url)  # loki cloud uses no suffix.
    g_headers = build_headers(api_key=config.api_key, additional_headers=config.headers)

    error_msg = ""
    for url in health_urls:
        try:
            _try_health_url(url, g_headers)
            return True, ""
        except Exception as e:
            logging.debug(
                f"Failed to fetch grafana health status at {url}", exc_info=True
            )
            error_msg += f"Failed to fetch grafana health status at {url}. {str(e)}\n"

    # Add helpful hint if this looks like a common misconfiguration
    if config.grafana_datasource_uid and ":3100" in config.url:
        error_msg += (
            "\n\nPossible configuration issue: grafana_datasource_uid is set but URL contains port 3100 "
            "(typically used for direct Loki connections). Please verify:\n"
            "- If connecting directly to Loki: remove grafana_datasource_uid from config\n"
            "- If connecting via Grafana proxy: ensure URL points to Grafana (usually port 3000)"
        )

    return False, error_msg



import os

import logging

from typing import Optional, Dict, Any, List
from holmes.common.env_vars import load_bool
from holmes.core.supabase_dal import SupabaseDal, FindingType
from holmes.core.tools import (
    StaticPrerequisite,
    Tool,
    ToolInvokeContext,
    ToolParameter,
    Toolset,
    ToolsetTag,
)
from holmes.core.tools import StructuredToolResult, StructuredToolResultStatus

PULL_EXTERNAL_FINDINGS = load_bool("PULL_EXTERNAL_FINDINGS", False)

PARAM_FINDING_ID = "id"
START_TIME = "start_datetime"
END_TIME = "end_datetime"
NAMESPACE = "namespace"
WORKLOAD = "workload"
DEFAULT_LIMIT_CHANGE_ROWS = 100
MAX_LIMIT_CHANGE_ROWS = 200


class FetchRobustaFinding(Tool):
    _dal: Optional[SupabaseDal]

    def __init__(self, dal: Optional[SupabaseDal]):
        super().__init__(
            name="fetch_finding_by_id",
            description="Fetches a robusta finding. Findings are events, like a Prometheus alert or a deployment update and configuration change.",
            parameters={
                PARAM_FINDING_ID: ToolParameter(
                    description="The id of the finding to fetch",
                    type="string",
                    required=True,
                )
            },
        )
        self._dal = dal

    def _fetch_finding(self, finding_id: str) -> Optional[Dict]:
        if self._dal and self._dal.enabled:
            return self._dal.get_issue_data(finding_id)
        else:
            error = f"Failed to find a finding with finding_id={finding_id}: Holmes' data access layer is not enabled."
            logging.error(error)
            return {"error": error}

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        finding_id = params[PARAM_FINDING_ID]
        try:
            finding = self._fetch_finding(finding_id)
            if finding:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.SUCCESS,
                    data=finding,
                    params=params,
                )
            else:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.NO_DATA,
                    data=f"Could not find a finding with finding_id={finding_id}",
                    params=params,
                )
        except Exception as e:
            logging.error(e)
            logging.error(
                f"There was an internal error while fetching finding {finding_id}. {str(e)}"
            )

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                data=f"There was an internal error while fetching finding {finding_id}",
                params=params,
            )

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return f"Robusta: Fetch finding data {params}"


class FetchResourceRecommendation(Tool):
    _dal: Optional[SupabaseDal]

    def __init__(self, dal: Optional[SupabaseDal]):
        super().__init__(
            name="fetch_resource_recommendation",
            description=(
                "Fetch KRR (Kubernetes Resource Recommendations) for CPU and memory optimization. "
                "KRR provides AI-powered recommendations based on actual historical usage patterns for right-sizing workloads. "
                "Supports two usage modes: "
                "(1) Specific workload lookup - Use name_pattern with an exact name, namespace, and kind to get recommendations for a single workload. "
                "(2) Discovery mode - Use limit and sort_by to get a ranked list of top optimization opportunities. Optionally filter by namespace, name_pattern (wildcards supported), kind, or container. "
                "Returns current configured resources alongside recommended values. In discovery mode, results are sorted by potential savings."
            ),
            parameters={
                "limit": ToolParameter(
                    description="Maximum number of recommendations to return (default: 10, max: 100).",
                    type="integer",
                    required=False,
                ),
                "sort_by": ToolParameter(
                    description=(
                        "Field to sort recommendations by potential savings. Options: "
                        "'cpu_total' (default) - Total CPU savings (requests + limits), "
                        "'memory_total' - Total memory savings (requests + limits), "
                        "'cpu_requests' - CPU requests savings, "
                        "'memory_requests' - Memory requests savings, "
                        "'cpu_limits' - CPU limits savings, "
                        "'memory_limits' - Memory limits savings, "
                        "'priority' - Use scan priority field."
                    ),
                    type="string",
                    required=False,
                ),
                "namespace": ToolParameter(
                    description="Filter by Kubernetes namespace (exact match). Leave empty to search all namespaces.",
                    type="string",
                    required=False,
                ),
                "name_pattern": ToolParameter(
                    description=(
                        "Filter by workload name pattern. Supports SQL LIKE patterns: "
                        "Use '%' as wildcard (e.g., '%app%' matches any name containing 'app', "
                        "'prod-%' matches names starting with 'prod-'). "
                        "Leave empty to match all names."
                    ),
                    type="string",
                    required=False,
                ),
                "kind": ToolParameter(
                    description=(
                        "Filter by Kubernetes resource kind. "
                        "Must be one of: Deployment, StatefulSet, DaemonSet, Job. "
                        "Leave empty to include all kinds."
                    ),
                    type="string",
                    required=False,
                ),
                "container": ToolParameter(
                    description="Filter by container name (exact match). Leave empty to include all containers.",
                    type="string",
                    required=False,
                ),
            },
        )
        self._dal = dal

    def _fetch_recommendations(self, params: Dict) -> Optional[List[Dict]]:
        if self._dal and self._dal.enabled:
            # Set default values
            limit = min(params.get("limit", 10) or 10, 100)
            sort_by = params.get("sort_by") or "cpu_total"

            return self._dal.get_resource_recommendation(
                limit=limit,
                sort_by=sort_by,
                namespace=params.get("namespace"),
                name_pattern=params.get("name_pattern"),
                kind=params.get("kind"),
                container=params.get("container"),
            )
        return None

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        try:
            recommendations = self._fetch_recommendations(params)
            if recommendations:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.SUCCESS,
                    data=recommendations,
                    params=params,
                )
            else:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.NO_DATA,
                    data=f"Could not find any recommendations with filters: {params}",
                    params=params,
                )
        except Exception as e:
            msg = f"There was an error while fetching top recommendations for {params}. {str(e)}"
            logging.exception(msg)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=msg,
                params=params,
            )

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return f"Robusta: Fetch KRR Recommendations ({str(params)})"


class FetchConfigurationChangesMetadataBase(Tool):
    _dal: Optional[SupabaseDal]

    def __init__(
        self,
        dal: Optional[SupabaseDal],
        name: str,
        description: str,
        add_cluster_filter: bool = True,
    ):
        """
        We need seperate tools for external and cluster configuration changes due to the different cluster parameters that are not on "external" changes like 'workload' and 'namespace'.
        add_cluster_filter: adds the namespace and workload parameters for configuration changes tool.
        """
        parameters = {
            START_TIME: ToolParameter(
                description="The starting time boundary for the search period. String in RFC3339 format.",
                type="string",
                required=True,
            ),
            END_TIME: ToolParameter(
                description="The ending time boundary for the search period. String in RFC3339 format.",
                type="string",
                required=True,
            ),
            "limit": ToolParameter(
                description=f"Maximum number of rows to return. Default is {DEFAULT_LIMIT_CHANGE_ROWS} and the maximum is 200",
                type="integer",
                required=False,
            ),
        }

        if add_cluster_filter:
            parameters.update(
                {
                    "namespace": ToolParameter(
                        description="The Kubernetes namespace name for filtering configuration changes",
                        type="string",
                        required=False,
                    ),
                    "workload": ToolParameter(
                        description="Kubernetes resource name to filter configuration changes (e.g., Pod, Deployment, Job, etc.). Must be the full name. For Pods, include the exact generated suffix.",
                        type="string",
                        required=False,
                    ),
                }
            )

        super().__init__(
            name=name,
            description=description,
            parameters=parameters,
        )
        self._dal = dal

    def _fetch_issues(
        self,
        params: Dict,
        cluster: Optional[str] = None,
        finding_type: FindingType = FindingType.CONFIGURATION_CHANGE,
    ) -> Optional[List[Dict]]:
        if self._dal and self._dal.enabled:
            return self._dal.get_issues_metadata(
                start_datetime=params["start_datetime"],
                end_datetime=params["end_datetime"],
                limit=min(
                    params.get("limit") or DEFAULT_LIMIT_CHANGE_ROWS,
                    MAX_LIMIT_CHANGE_ROWS,
                ),
                ns=params.get("namespace"),
                workload=params.get("workload"),
                cluster=cluster,
                finding_type=finding_type,
            )
        return None

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        try:
            changes = self._fetch_issues(params)
            if changes:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.SUCCESS,
                    data=changes,
                    params=params,
                )
            else:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.NO_DATA,
                    data=f"{self.name} found no data. {params}",
                    params=params,
                )
        except Exception as e:
            msg = f"There was an internal error while fetching changes for {params}. {str(e)}"
            logging.exception(msg)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                data=msg,
                params=params,
            )

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return f"Robusta: Search Change History {params}"


class FetchConfigurationChangesMetadata(FetchConfigurationChangesMetadataBase):
    def __init__(self, dal: Optional[SupabaseDal]):
        super().__init__(
            dal=dal,
            name="fetch_configuration_changes_metadata",
            description=(
                "Fetch configuration changes metadata in a given time range. "
                "By default, fetch all cluster changes. Can be filtered on a given namespace or a specific kubernetes resource. "
                "Use fetch_finding_by_id to get detailed change of one specific configuration change."
            ),
        )


class FetchExternalConfigurationChangesMetadata(FetchConfigurationChangesMetadataBase):
    """
    Fetch configuration changes from external sources, e.g., LaunchDarkly changes.
    It needs to be a seperate tool due to the different cluster parameter used in the DAL method like workload and namespace.
    """

    def __init__(self, dal: Optional[SupabaseDal]):
        super().__init__(
            dal=dal,
            name="fetch_external_configuration_changes_metadata",
            description=(
                "Fetch external configuration changes metadata in a given time range. "
                "Fetches configuration changes from external sources. "
                "Use fetch_finding_by_id to get detailed change of one specific configuration change."
            ),
            add_cluster_filter=False,
        )

    def _fetch_issues(self, params: Dict) -> Optional[List[Dict]]:  # type: ignore
        return super()._fetch_issues(params, cluster="external")

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return f"Robusta: Search External Change History {params}"


class FetchResourceIssuesMetadata(FetchConfigurationChangesMetadataBase):
    def __init__(self, dal: Optional[SupabaseDal]):
        super().__init__(
            dal=dal,
            name="fetch_resource_issues_metadata",
            description=(
                "Fetch issues and alert metadata in a given time range. "
                "Must be filtered on a given namespace and specific kubernetes resource, such as pod, deployment, job, etc. "
                "Use fetch_finding_by_id to get further information on a specific issue or alert."
            ),
            add_cluster_filter=False,
        )
        self.parameters.update(
            {
                "namespace": ToolParameter(
                    description="The Kubernetes namespace name for filtering issues and alerts",
                    type="string",
                    required=True,
                ),
                "workload": ToolParameter(
                    description="Kubernetes resource name to filter issues and alerts (e.g., Pod, Deployment, Job, etc.). Must be the full name. For Pods, include the exact generated suffix.",
                    type="string",
                    required=True,
                ),
            }
        )

    def _fetch_issues(self, params: Dict) -> Optional[List[Dict]]:  # type: ignore
        return super()._fetch_issues(params, finding_type=FindingType.ISSUE)

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return f"Robusta: fetch resource issues metadata {params}"


class RobustaToolset(Toolset):
    def __init__(self, dal: Optional[SupabaseDal]):
        dal_prereq = StaticPrerequisite(
            enabled=True if dal else False,
            disabled_reason="Integration with Robusta cloud is disabled",
        )
        if dal:
            dal_prereq = StaticPrerequisite(
                enabled=dal.enabled, disabled_reason="Data access layer is disabled"
            )

        tools = [
            FetchRobustaFinding(dal),
            FetchConfigurationChangesMetadata(dal),
            FetchResourceRecommendation(dal),
            FetchResourceIssuesMetadata(dal),
        ]

        if PULL_EXTERNAL_FINDINGS:
            tools.append(FetchExternalConfigurationChangesMetadata(dal))

        super().__init__(
            icon_url="https://cdn.prod.website-files.com/633e9bac8f71dfb7a8e4c9a6/646be7710db810b14133bdb5_logo.svg",
            description="Fetches alerts metadata and change history",
            docs_url="https://holmesgpt.dev/data-sources/builtin-toolsets/robusta/",
            name="robusta",
            prerequisites=[dal_prereq],
            tools=tools,
            tags=[
                ToolsetTag.CORE,
            ],
            is_default=True,
        )
        template_file_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "robusta_instructions.jinja2")
        )
        self._load_llm_instructions(jinja_template=f"file://{template_file_path}")

    def get_example_config(self) -> Dict[str, Any]:
        return {}



import json
import logging
import os
import threading
from abc import abstractmethod
from math import floor
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

import litellm
from litellm.litellm_core_utils.streaming_handler import CustomStreamWrapper
from litellm.types.utils import ModelResponse, TextCompletionResponse
import sentry_sdk
from pydantic import BaseModel, ConfigDict, SecretStr
from typing_extensions import Self

from holmes.clients.robusta_client import (
    RobustaModel,
    RobustaModelsResponse,
    fetch_robusta_models,
)

from holmes.common.env_vars import (
    FALLBACK_CONTEXT_WINDOW_SIZE,
    LOAD_ALL_ROBUSTA_MODELS,
    REASONING_EFFORT,
    ROBUSTA_AI,
    ROBUSTA_API_ENDPOINT,
    THINKING,
    EXTRA_HEADERS,
    TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT,
    TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_TOKENS,
)
from holmes.core.supabase_dal import SupabaseDal
from holmes.utils.env import environ_get_safe_int, replace_env_vars_values
from holmes.utils.file_utils import load_yaml_file

if TYPE_CHECKING:
    from holmes.config import Config

MODEL_LIST_FILE_LOCATION = os.environ.get(
    "MODEL_LIST_FILE_LOCATION", "/etc/holmes/config/model_list.yaml"
)


OVERRIDE_MAX_OUTPUT_TOKEN = environ_get_safe_int("OVERRIDE_MAX_OUTPUT_TOKEN")
OVERRIDE_MAX_CONTENT_SIZE = environ_get_safe_int("OVERRIDE_MAX_CONTENT_SIZE")


def get_context_window_compaction_threshold_pct() -> int:
    """Get the compaction threshold percentage at runtime to support test overrides."""
    return environ_get_safe_int("CONTEXT_WINDOW_COMPACTION_THRESHOLD_PCT", default="95")


ROBUSTA_AI_MODEL_NAME = "Robusta"


class TokenCountMetadata(BaseModel):
    total_tokens: int
    tools_tokens: int
    system_tokens: int
    user_tokens: int
    tools_to_call_tokens: int
    assistant_tokens: int
    other_tokens: int


class ModelEntry(BaseModel):
    """ModelEntry represents a single LLM model configuration."""

    model: str
    # TODO: the name field seems to be redundant, can we remove it?
    name: Optional[str] = None
    api_key: Optional[SecretStr] = None
    base_url: Optional[str] = None
    is_robusta_model: Optional[bool] = None
    custom_args: Optional[Dict[str, Any]] = None

    # LLM configurations used services like Azure OpenAI Service
    api_base: Optional[str] = None
    api_version: Optional[str] = None

    model_config = ConfigDict(
        extra="allow",
    )

    @classmethod
    def load_from_dict(cls, data: dict) -> Self:
        return cls.model_validate(data)


class LLM:
    @abstractmethod
    def __init__(self):
        self.model: str  # type: ignore

    @abstractmethod
    def get_context_window_size(self) -> int:
        pass

    @abstractmethod
    def get_maximum_output_token(self) -> int:
        pass

    def get_max_token_count_for_single_tool(self) -> int:
        if (
            0 < TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT
            and TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT <= 100
        ):
            context_window_size = self.get_context_window_size()
            calculated_max_tokens = int(
                context_window_size * TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_PCT // 100
            )
            return min(calculated_max_tokens, TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_TOKENS)
        else:
            return TOOL_MAX_ALLOCATED_CONTEXT_WINDOW_TOKENS

    @abstractmethod
    def count_tokens(
        self, messages: list[dict], tools: Optional[list[dict[str, Any]]] = None
    ) -> TokenCountMetadata:
        pass

    @abstractmethod
    def completion(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = [],
        tool_choice: Optional[Union[str, dict]] = None,
        response_format: Optional[Union[dict, Type[BaseModel]]] = None,
        temperature: Optional[float] = None,
        drop_params: Optional[bool] = None,
        stream: Optional[bool] = None,
    ) -> Union[ModelResponse, CustomStreamWrapper]:
        pass


class DefaultLLM(LLM):
    model: str
    api_key: Optional[str]
    api_base: Optional[str]
    api_version: Optional[str]
    args: Dict
    is_robusta_model: bool

    def __init__(
        self,
        model: str,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        args: Optional[Dict] = None,
        tracer: Optional[Any] = None,
        name: Optional[str] = None,
        is_robusta_model: bool = False,
    ):
        self.model = model
        self.api_key = api_key
        self.api_base = api_base
        self.api_version = api_version
        self.args = args or {}
        self.tracer = tracer
        self.name = name
        self.is_robusta_model = is_robusta_model
        self.update_custom_args()
        self.check_llm(
            self.model, self.api_key, self.api_base, self.api_version, self.args
        )

    def update_custom_args(self):
        self.max_context_size = self.args.get("custom_args", {}).get("max_context_size")
        self.args.pop("custom_args", None)

    def check_llm(
        self,
        model: str,
        api_key: Optional[str],
        api_base: Optional[str],
        api_version: Optional[str],
        args: Optional[dict] = None,
    ):
        if self.is_robusta_model:
            # The model is assumed correctly configured if it is a robusta model
            # For robusta models, this code would fail because Holmes has no knowledge of the API keys
            # to azure or bedrock as all completion API calls go through robusta's LLM proxy
            return
        args = args or {}
        logging.debug(f"Checking LiteLLM model {model}")
        lookup = litellm.get_llm_provider(model)
        if not lookup:
            raise Exception(f"Unknown provider for model {model}")
        provider = lookup[1]
        if provider == "watsonx":
            # NOTE: LiteLLM's validate_environment does not currently include checks for IBM WatsonX.
            # The following WatsonX-specific variables are set based on documentation from:
            # https://docs.litellm.ai/docs/providers/watsonx
            # Required variables for WatsonX:
            # - WATSONX_URL: Base URL of your WatsonX instance (required)
            # - WATSONX_APIKEY or WATSONX_TOKEN: IBM Cloud API key or IAM auth token (one is required)
            model_requirements = {"missing_keys": [], "keys_in_environment": True}
            if api_key:
                os.environ["WATSONX_APIKEY"] = api_key
            if "WATSONX_URL" not in os.environ:
                model_requirements["missing_keys"].append("WATSONX_URL")  # type: ignore
                model_requirements["keys_in_environment"] = False
            if "WATSONX_APIKEY" not in os.environ and "WATSONX_TOKEN" not in os.environ:
                model_requirements["missing_keys"].extend(  # type: ignore
                    ["WATSONX_APIKEY", "WATSONX_TOKEN"]
                )
                model_requirements["keys_in_environment"] = False
            # WATSONX_PROJECT_ID is required because we don't let user pass it to completion call directly
            if "WATSONX_PROJECT_ID" not in os.environ:
                model_requirements["missing_keys"].append("WATSONX_PROJECT_ID")  # type: ignore
                model_requirements["keys_in_environment"] = False
            # https://docs.litellm.ai/docs/providers/watsonx#usage---models-in-deployment-spaces
            # using custom watsonx deployments might require to set WATSONX_DEPLOYMENT_SPACE_ID env
            if "watsonx/deployment/" in self.model:
                logging.warning(
                    "Custom WatsonX deployment detected. You may need to set the WATSONX_DEPLOYMENT_SPACE_ID "
                    "environment variable for proper functionality. For more information, refer to the documentation: "
                    "https://docs.litellm.ai/docs/providers/watsonx#usage---models-in-deployment-spaces"
                )
        elif provider == "bedrock":
            if os.environ.get("AWS_PROFILE") or os.environ.get(
                "AWS_BEARER_TOKEN_BEDROCK"
            ):
                model_requirements = {"keys_in_environment": True, "missing_keys": []}
            elif args.get("aws_access_key_id") and args.get("aws_secret_access_key"):
                return  # break fast.
            else:
                model_requirements = litellm.validate_environment(
                    model=model, api_key=api_key, api_base=api_base
                )
        else:
            model_requirements = litellm.validate_environment(
                model=model, api_key=api_key, api_base=api_base
            )
            # validate_environment does not accept api_version, and as a special case for Azure OpenAI Service,
            # when all the other AZURE environments are set expect AZURE_API_VERSION, validate_environment complains
            # the missing of it even after the api_version is set.
            # TODO: There's an open PR in litellm to accept api_version in validate_environment, we can leverage this
            # change if accepted to ignore the following check.
            # https://github.com/BerriAI/litellm/pull/13808
            if (
                provider == "azure"
                and ["AZURE_API_VERSION"] == model_requirements["missing_keys"]
                and api_version is not None
            ):
                model_requirements["missing_keys"] = []
                model_requirements["keys_in_environment"] = True

        if not model_requirements["keys_in_environment"]:
            raise Exception(
                f"model {model} requires the following environment variables: {model_requirements['missing_keys']}"
            )

    def _get_model_name_variants_for_lookup(self) -> list[str]:
        """
        Generate model name variants to try when looking up in litellm.model_cost.
        Returns a list of names to try in order: exact, lowercase, without prefix, etc.
        """
        names_to_try = [self.model, self.model.lower()]

        # If there's a prefix, also try without it
        if "/" in self.model:
            base_model = self.model.split("/", 1)[1]
            names_to_try.extend([base_model, base_model.lower()])

        # Remove duplicates while preserving order (dict.fromkeys maintains insertion order in Python 3.7+)
        return list(dict.fromkeys(names_to_try))

    def get_context_window_size(self) -> int:
        if self.max_context_size:
            return self.max_context_size

        if OVERRIDE_MAX_CONTENT_SIZE:
            logging.debug(
                f"Using override OVERRIDE_MAX_CONTENT_SIZE {OVERRIDE_MAX_CONTENT_SIZE}"
            )
            return OVERRIDE_MAX_CONTENT_SIZE

        # Try each name variant
        for name in self._get_model_name_variants_for_lookup():
            try:
                return litellm.model_cost[name]["max_input_tokens"]
            except Exception:
                continue

        # Log which lookups we tried
        logging.warning(
            f"Couldn't find model {self.model} in litellm's model list (tried: {', '.join(self._get_model_name_variants_for_lookup())}), "
            f"using default {FALLBACK_CONTEXT_WINDOW_SIZE} tokens for max_input_tokens. "
            f"To override, set OVERRIDE_MAX_CONTENT_SIZE environment variable to the correct value for your model."
        )
        return FALLBACK_CONTEXT_WINDOW_SIZE

    @sentry_sdk.trace
    def count_tokens(
        self, messages: list[dict], tools: Optional[list[dict[str, Any]]] = None
    ) -> TokenCountMetadata:
        # TODO: Add a recount:bool flag to save time. When the flag is false, reuse 'message["token_count"]' for individual messages.
        # It's only necessary to recount message tokens at the beginning of a session because the LLM model may have changed.
        # Changing the model requires recounting tokens because the tokenizer may be different
        total_tokens = 0
        tools_tokens = 0
        system_tokens = 0
        assistant_tokens = 0
        user_tokens = 0
        other_tokens = 0
        tools_to_call_tokens = 0
        for message in messages:
            # count message tokens individually because it gives us fine grain information about each tool call/message etc.
            # However be aware that the sum of individual message tokens is not equal to the overall messages token
            token_count = litellm.token_counter(  # type: ignore
                model=self.model, messages=[message]
            )
            message["token_count"] = token_count
            role = message.get("role")
            if role == "system":
                system_tokens += token_count
            elif role == "user":
                user_tokens += token_count
            elif role == "tool":
                tools_tokens += token_count
            elif role == "assistant":
                assistant_tokens += token_count
            else:
                # although this should not be needed,
                # it is defensive code so that all tokens are accounted for
                # and can potentially make debugging easier
                other_tokens += token_count

        messages_token_count_without_tools = litellm.token_counter(  # type: ignore
            model=self.model, messages=messages
        )

        total_tokens = litellm.token_counter(  # type: ignore
            model=self.model,
            messages=messages,
            tools=tools,  # type: ignore
        )
        tools_to_call_tokens = max(0, total_tokens - messages_token_count_without_tools)

        return TokenCountMetadata(
            total_tokens=total_tokens,
            system_tokens=system_tokens,
            user_tokens=user_tokens,
            tools_tokens=tools_tokens,
            tools_to_call_tokens=tools_to_call_tokens,
            other_tokens=other_tokens,
            assistant_tokens=assistant_tokens,
        )

    def get_litellm_corrected_name_for_robusta_ai(self) -> str:
        if self.is_robusta_model:
            # For robusta models, self.model is the underlying provider/model used by Robusta AI
            # To avoid litellm modifying the API URL according to the provider, the provider name
            # is replaced with 'openai/' just before doing a completion() call
            # Cf. https://docs.litellm.ai/docs/providers/openai_compatible
            split_model_name = self.model.split("/")
            return (
                split_model_name[0]
                if len(split_model_name) == 1
                else f"openai/{split_model_name[1]}"
            )
        else:
            return self.model

    def completion(
        self,
        messages: List[Dict[str, Any]],
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, dict]] = None,
        response_format: Optional[Union[dict, Type[BaseModel]]] = None,
        temperature: Optional[float] = None,
        drop_params: Optional[bool] = None,
        stream: Optional[bool] = None,
    ) -> Union[ModelResponse, CustomStreamWrapper]:
        tools_args = {}
        allowed_openai_params = None

        if tools and len(tools) > 0 and tool_choice == "auto":
            tools_args["tools"] = tools
            tools_args["tool_choice"] = tool_choice  # type: ignore

        if THINKING:
            self.args.setdefault("thinking", json.loads(THINKING))

        if EXTRA_HEADERS:
            self.args.setdefault("extra_headers", json.loads(EXTRA_HEADERS))

        if self.args.get("thinking", None):
            litellm.modify_params = True

        if REASONING_EFFORT:
            self.args.setdefault("reasoning_effort", REASONING_EFFORT)
            allowed_openai_params = [
                "reasoning_effort"
            ]  # can be removed after next litelm version

        existing_allowed = self.args.pop("allowed_openai_params", None)
        if existing_allowed:
            if allowed_openai_params is None:
                allowed_openai_params = []
            allowed_openai_params.extend(existing_allowed)

        self.args.setdefault("temperature", temperature)

        self._add_cache_control_to_last_message(messages)

        # Get the litellm module to use (wrapped or unwrapped)
        litellm_to_use = self.tracer.wrap_llm(litellm) if self.tracer else litellm

        litellm_model_name = self.get_litellm_corrected_name_for_robusta_ai()
        result = litellm_to_use.completion(
            model=litellm_model_name,
            api_key=self.api_key,
            base_url=self.api_base,
            api_version=self.api_version,
            messages=messages,
            response_format=response_format,
            drop_params=drop_params,
            allowed_openai_params=allowed_openai_params,
            stream=stream,
            **tools_args,
            **self.args,
        )

        if isinstance(result, ModelResponse):
            return result
        elif isinstance(result, CustomStreamWrapper):
            return result
        else:
            raise Exception(f"Unexpected type returned by the LLM {type(result)}")

    def get_maximum_output_token(self) -> int:
        max_output_tokens = floor(min(64000, self.get_context_window_size() / 5))

        if OVERRIDE_MAX_OUTPUT_TOKEN:
            logging.debug(
                f"Using OVERRIDE_MAX_OUTPUT_TOKEN {OVERRIDE_MAX_OUTPUT_TOKEN}"
            )
            return OVERRIDE_MAX_OUTPUT_TOKEN

        # Try each name variant
        for name in self._get_model_name_variants_for_lookup():
            try:
                litellm_max_output_tokens = litellm.model_cost[name][
                    "max_output_tokens"
                ]
                if litellm_max_output_tokens < max_output_tokens:
                    max_output_tokens = litellm_max_output_tokens
                return max_output_tokens
            except Exception:
                continue

        # Log which lookups we tried
        logging.warning(
            f"Couldn't find model {self.model} in litellm's model list (tried: {', '.join(self._get_model_name_variants_for_lookup())}), "
            f"using {max_output_tokens} tokens for max_output_tokens. "
            f"To override, set OVERRIDE_MAX_OUTPUT_TOKEN environment variable to the correct value for your model."
        )
        return max_output_tokens

    def _add_cache_control_to_last_message(
        self, messages: List[Dict[str, Any]]
    ) -> None:
        """
        Add cache_control to the last non-user message for Anthropic prompt caching.
        Removes any existing cache_control from previous messages to avoid accumulation.
        """
        # Skip cache_control for VertexAI/Gemini models as they don't support it with tools
        if self.model and (
            "vertex" in self.model.lower() or "gemini" in self.model.lower()
        ):
            return

        # First, remove any existing cache_control from all messages
        for msg in messages:
            content = msg.get("content")
            if isinstance(content, list):
                for block in content:
                    if isinstance(block, dict) and "cache_control" in block:
                        del block["cache_control"]
                        logging.debug(
                            f"Removed existing cache_control from {msg.get('role')} message"
                        )

        # Find the last non-user message to add cache_control to.
        # Adding cache_control to user message requires changing its structure, so we avoid it
        # This avoids breaking parse_messages_tags which only processes user messages
        target_msg = None
        for msg in reversed(messages):
            if msg.get("role") != "user":
                target_msg = msg
                break

        if not target_msg:
            logging.debug("No non-user message found for cache_control")
            return

        content = target_msg.get("content")

        if content is None:
            return

        if isinstance(content, str) and content:
            # Convert string to structured format with cache_control
            target_msg["content"] = [
                {
                    "type": "text",
                    "text": content,
                    "cache_control": {"type": "ephemeral"},
                }
            ]
            logging.debug(
                f"Added cache_control to {target_msg.get('role')} message (converted from string)"
            )
        elif isinstance(content, list) and content:
            # Add cache_control to the last content block
            last_block = content[-1]
            if isinstance(last_block, dict) and "type" in last_block:
                last_block["cache_control"] = {"type": "ephemeral"}
                logging.debug(
                    f"Added cache_control to {target_msg.get('role')} message (structured content)"
                )


class LLMModelRegistry:
    def __init__(self, config: "Config", dal: SupabaseDal) -> None:
        self.config = config
        self._llms: dict[str, ModelEntry] = {}
        self._default_robusta_model = None
        self.dal = dal
        self._lock = threading.RLock()

        self._init_models()

    @property
    def default_robusta_model(self) -> Optional[str]:
        return self._default_robusta_model

    def _init_models(self):
        self._llms = self._parse_models_file(MODEL_LIST_FILE_LOCATION)

        if self._should_load_robusta_ai():
            self.configure_robusta_ai_model()

        if self._should_load_config_model():
            self._llms[self.config.model] = self._create_model_entry(
                model=self.config.model,
                model_name=self.config.model,
                base_url=self.config.api_base,
                is_robusta_model=False,
                api_key=self.config.api_key,
                api_version=self.config.api_version,
            )

    def _should_load_config_model(self) -> bool:
        if self.config.model is not None:
            if self._llms and self.config.model in self._llms:
                # model already loaded from file
                return False
            return True

        # backward compatibility - in the past config.model was set by default to gpt-4o.
        # so we need to check if the user has set an OPENAI_API_KEY to load the config model.
        has_openai_key = os.environ.get("OPENAI_API_KEY")
        if has_openai_key:
            self.config.model = "gpt-4.1"
            return True

        return False

    def configure_robusta_ai_model(self) -> None:
        try:
            if not self.config.cluster_name or not LOAD_ALL_ROBUSTA_MODELS:
                self._load_default_robusta_config()
                return

            if not self.dal.account_id or not self.dal.enabled:
                self._load_default_robusta_config()
                return

            account_id, token = self.dal.get_ai_credentials()

            robusta_models: RobustaModelsResponse | None = fetch_robusta_models(
                account_id, token
            )
            if not robusta_models or not robusta_models.models:
                self._load_default_robusta_config()
                return

            default_model = None
            for model_name, model_data in robusta_models.models.items():
                logging.info(f"Loading Robusta AI model: {model_name}")
                self._llms[model_name] = self._create_robusta_model_entry(
                    model_name=model_name, model_data=model_data
                )
                if model_data.is_default:
                    default_model = model_name

            if default_model:
                logging.info(f"Setting default Robusta AI model to: {default_model}")
                self._default_robusta_model: str = default_model  # type: ignore

        except Exception:
            logging.exception("Failed to get all robusta models")
            # fallback to default behavior
            self._load_default_robusta_config()

    def _load_default_robusta_config(self):
        if self._should_load_robusta_ai():
            logging.info("Loading default Robusta AI model")
            self._llms[ROBUSTA_AI_MODEL_NAME] = ModelEntry(
                name=ROBUSTA_AI_MODEL_NAME,
                model="gpt-4o",  # TODO: tech debt, this isn't really
                base_url=ROBUSTA_API_ENDPOINT,
                is_robusta_model=True,
            )
            self._default_robusta_model = ROBUSTA_AI_MODEL_NAME

    def _should_load_robusta_ai(self) -> bool:
        if not self.config.should_try_robusta_ai:
            return False

        # ROBUSTA_AI were set in the env vars, so we can use it directly
        if ROBUSTA_AI is not None:
            return ROBUSTA_AI

        # MODEL is set in the env vars, e.g. the user is using a custom model
        # so we don't need to load the robusta AI model and keep the behavior backward compatible
        if "MODEL" in os.environ:
            return False

        # if the user has provided a model list, we don't need to load the robusta AI model
        if self._llms:
            return False

        return True

    def get_model_params(self, model_key: Optional[str] = None) -> ModelEntry:
        with self._lock:
            if not self._llms:
                raise Exception("No llm models were loaded")

            if model_key:
                model_params = self._llms.get(model_key)
                if model_params:
                    logging.info(f"Using selected model: {model_key}")
                    return model_params.model_copy()

                if model_key.startswith("Robusta/"):
                    logging.warning("Resyncing Registry and Robusta models.")
                    self._init_models()
                    model_params = self._llms.get(model_key)
                    if model_params:
                        logging.info(f"Using selected model: {model_key}")
                        return model_params.model_copy()

                logging.error(f"Couldn't find model: {model_key} in model list")

            if self._default_robusta_model:
                model_params = self._llms.get(self._default_robusta_model)
                if model_params is not None:
                    logging.info(
                        f"Using default Robusta AI model: {self._default_robusta_model}"
                    )
                    return model_params.model_copy()

                logging.error(
                    f"Couldn't find default Robusta AI model: {self._default_robusta_model} in model list"
                )

            model_key, first_model_params = next(iter(self._llms.items()))
            logging.debug(f"Using first available model: {model_key}")
            return first_model_params.model_copy()

    @property
    def models(self) -> dict[str, ModelEntry]:
        with self._lock:
            return self._llms

    def _parse_models_file(self, path: str) -> dict[str, ModelEntry]:
        models = load_yaml_file(path, raise_error=False, warn_not_found=False)
        for _, params in models.items():
            params = replace_env_vars_values(params)

        llms = {}
        for model_name, params in models.items():
            llms[model_name] = ModelEntry.model_validate(params)

        return llms

    def _create_robusta_model_entry(
        self, model_name: str, model_data: RobustaModel
    ) -> ModelEntry:
        entry = self._create_model_entry(
            model=model_data.model,
            model_name=model_name,
            base_url=f"{ROBUSTA_API_ENDPOINT}/llm/{model_name}",
            is_robusta_model=True,
        )
        entry.custom_args = model_data.holmes_args or {}  # type: ignore[assignment]
        return entry

    def _create_model_entry(
        self,
        model: str,
        model_name: str,
        base_url: Optional[str] = None,
        is_robusta_model: Optional[bool] = None,
        api_key: Optional[SecretStr] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
    ) -> ModelEntry:
        return ModelEntry(
            name=model_name,
            model=model,
            base_url=base_url,
            is_robusta_model=is_robusta_model,
            api_key=api_key,
            api_base=api_base,
            api_version=api_version,
        )


def get_llm_usage(
    llm_response: Union[ModelResponse, CustomStreamWrapper, TextCompletionResponse],
) -> dict:
    usage: dict = {}
    if (
        (
            isinstance(llm_response, ModelResponse)
            or isinstance(llm_response, TextCompletionResponse)
        )
        and hasattr(llm_response, "usage")
        and llm_response.usage
    ):  # type: ignore
        usage["prompt_tokens"] = llm_response.usage.prompt_tokens  # type: ignore
        usage["completion_tokens"] = llm_response.usage.completion_tokens  # type: ignore
        usage["total_tokens"] = llm_response.usage.total_tokens  # type: ignore
    elif isinstance(llm_response, CustomStreamWrapper):
        complete_response = litellm.stream_chunk_builder(chunks=llm_response)  # type: ignore
        if complete_response:
            return get_llm_usage(complete_response)
    return usage



import logging
import os
from typing import Any, Dict

from uuid import uuid4

from holmes.core.todo_tasks_formatter import format_tasks
from holmes.core.tools import (
    StructuredToolResult,
    StructuredToolResultStatus,
    Tool,
    ToolInvokeContext,
    ToolParameter,
    Toolset,
    ToolsetTag,
)
from holmes.plugins.toolsets.investigator.model import Task, TaskStatus

TODO_WRITE_TOOL_NAME = "TodoWrite"


def parse_tasks(todos_data: Any) -> list[Task]:
    tasks = []

    for todo_item in todos_data:
        if isinstance(todo_item, dict):
            task = Task(
                id=todo_item.get("id", str(uuid4())),
                content=todo_item.get("content", ""),
                status=TaskStatus(todo_item.get("status", "pending")),
            )
            tasks.append(task)

    return tasks


class TodoWriteTool(Tool):
    name: str = TODO_WRITE_TOOL_NAME
    description: str = "Save investigation tasks to break down complex problems into manageable sub-tasks. ALWAYS provide the COMPLETE list of all tasks, not just the ones being updated."
    parameters: Dict[str, ToolParameter] = {
        "todos": ToolParameter(
            description="COMPLETE list of ALL tasks on the task list. Each task should have: id (string), content (string), status (pending/in_progress/completed/failed)",
            type="array",
            required=True,
            items=ToolParameter(
                type="object",
                properties={
                    "id": ToolParameter(type="string", required=True),
                    "content": ToolParameter(type="string", required=True),
                    "status": ToolParameter(
                        type="string",
                        required=True,
                        enum=["pending", "in_progress", "completed", "failed"],
                    ),
                },
            ),
        ),
    }

    # Print a nice table to console/log
    def print_tasks_table(self, tasks):
        if not tasks:
            logging.info("No tasks in the investigation plan.")
            return

        status_icons = {
            "pending": "[ ]",
            "in_progress": "[~]",
            "completed": "[✓]",
            "failed": "[✗]",
        }

        max_id_width = max(len(str(task.id)) for task in tasks)
        max_content_width = max(len(task.content) for task in tasks)
        max_status_display_width = max(
            len(f"{status_icons[task.status.value]} {task.status.value}")
            for task in tasks
        )

        id_width = max(max_id_width, len("ID"))
        content_width = max(max_content_width, len("Content"))
        status_width = max(max_status_display_width, len("Status"))

        separator = f"+{'-' * (id_width + 2)}+{'-' * (content_width + 2)}+{'-' * (status_width + 2)}+"
        header = f"| {'ID':<{id_width}} | {'Content':<{content_width}} | {'Status':<{status_width}} |"
        tasks_to_display = []

        for task in tasks:
            status_display = f"{status_icons[task.status.value]} {task.status.value}"
            row = f"| {task.id:<{id_width}} | {task.content:<{content_width}} | {status_display:<{status_width}} |"
            tasks_to_display.append(row)

        logging.info(
            f"Task List:\n{separator}\n{header}\n{separator}\n"
            + "\n".join(tasks_to_display)
            + f"\n{separator}"
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        try:
            todos_data = params.get("todos", [])

            tasks = parse_tasks(todos_data=todos_data)

            logging.debug(f"Tasks: {len(tasks)}")

            self.print_tasks_table(tasks)
            formatted_tasks = format_tasks(tasks)

            response_data = f"✅ Investigation plan updated with {len(tasks)} tasks. Tasks are now stored in session and will appear in subsequent prompts.\n\n"
            if formatted_tasks:
                response_data += formatted_tasks
            else:
                response_data += "No tasks currently in the investigation plan."

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=response_data,
                params=params,
            )

        except Exception as e:
            logging.exception("error using todowrite tool")
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Failed to process tasks: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params: Dict) -> str:
        return "Update investigation tasks"


class CoreInvestigationToolset(Toolset):
    """Core toolset for investigation management and task planning."""

    def __init__(self):
        super().__init__(
            name="core_investigation",
            description="Core investigation tools for task management and planning",
            enabled=True,
            tools=[TodoWriteTool()],
            tags=[ToolsetTag.CORE],
            is_default=True,
        )

    def get_example_config(self) -> Dict[str, Any]:
        return {}

    def _reload_instructions(self):
        template_file_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "investigator_instructions.jinja2")
        )
        self._load_llm_instructions(jinja_template=f"file://{template_file_path}")



import logging
from typing import Any, Dict, Tuple
from datetime import datetime, timezone

from holmes.core.tools import (
    StructuredToolResult,
    ToolInvokeContext,
    ToolParameter,
    StructuredToolResultStatus,
)
from holmes.plugins.toolsets.azure_sql.azure_base_toolset import (
    BaseAzureSQLTool,
    BaseAzureSQLToolset,
    AzureSQLDatabaseConfig,
)
from holmes.plugins.toolsets.azure_sql.apis.storage_analysis_api import (
    StorageAnalysisAPI,
)
from holmes.plugins.toolsets.azure_sql.apis.azure_sql_api import AzureSQLAPIClient
from holmes.plugins.toolsets.utils import toolset_name_for_one_liner


class AnalyzeDatabaseStorage(BaseAzureSQLTool):
    def __init__(self, toolset: "BaseAzureSQLToolset"):
        super().__init__(
            name="analyze_database_storage",
            description="Analyzes database storage utilization including disk usage, growth trends, file-level details, and table space consumption. Use this for capacity planning and storage optimization.",
            parameters={
                "hours_back": ToolParameter(
                    description="Time window for storage metrics analysis in hours. Use 24 for daily trends, 168 for weekly analysis (default: 24)",
                    type="integer",
                    required=False,
                ),
                "top_tables": ToolParameter(
                    description="Number of largest tables to analyze for space usage. Use 20 for comprehensive view, 10 for quick overview (default: 20)",
                    type="integer",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _build_storage_report(
        self,
        db_config: AzureSQLDatabaseConfig,
        storage_data: Dict,
        hours_back: int,
        top_tables: int,
    ) -> str:
        """Build the formatted storage report from gathered data."""
        report_sections = []

        # Header
        report_sections.append("# Azure SQL Database Storage Analysis Report")
        report_sections.append(f"**Database:** {db_config.database_name}")
        report_sections.append(f"**Server:** {db_config.server_name}")
        report_sections.append(f"**Analysis Period:** Last {hours_back} hours")
        report_sections.append(
            f"**Generated:** {datetime.now(timezone.utc).isoformat()}"
        )
        report_sections.append("")

        # Storage Summary
        report_sections.append("## Storage Summary")
        summary = storage_data.get("summary", {})
        if "error" in summary:
            report_sections.append(
                f"⚠️ **Error retrieving storage summary:** {summary['error']}"
            )
        else:
            total_size = summary.get("total_database_size_mb", 0) or 0
            used_size = summary.get("total_used_size_mb", 0) or 0
            data_size = summary.get("total_data_size_mb", 0) or 0
            log_size = summary.get("total_log_size_mb", 0) or 0

            if total_size:
                used_percent = (used_size / total_size) * 100
                free_size = total_size - used_size

                report_sections.append(
                    f"- **Total Database Size**: {total_size:,.1f} MB"
                )
                report_sections.append(
                    f"- **Used Space**: {used_size:,.1f} MB ({used_percent:.1f}%)"
                )
                report_sections.append(f"- **Free Space**: {free_size:,.1f} MB")
                report_sections.append(f"- **Data Files**: {data_size:,.1f} MB")
                report_sections.append(f"- **Log Files**: {log_size:,.1f} MB")
                report_sections.append(
                    f"- **Data Files Count**: {summary.get('data_files_count', 0)}"
                )
                report_sections.append(
                    f"- **Log Files Count**: {summary.get('log_files_count', 0)}"
                )
            else:
                report_sections.append("No storage summary data available")
        report_sections.append("")

        # File Details
        report_sections.append("## Database Files Details")
        file_details = storage_data.get("file_details", [])
        if isinstance(file_details, dict) and "error" in file_details:
            report_sections.append(
                f"⚠️ **Error retrieving file details:** {file_details['error']}"
            )
        elif file_details:
            for file_info in file_details:
                file_type = file_info.get("file_type", "Unknown")
                logical_name = file_info.get("logical_name", "Unknown")
                size_mb = file_info.get("size_mb", 0) or 0
                used_mb = file_info.get("used_mb")
                used_percent = file_info.get("used_percent")
                max_size = file_info.get("max_size", "Unknown")
                growth = file_info.get("growth_setting", "Unknown")

                # Only calculate status icon if we have used_percent data
                if used_percent is not None:
                    status_icon = (
                        "🔴"
                        if used_percent > 90
                        else "🟡"
                        if used_percent > 75
                        else "🟢"
                    )
                else:
                    status_icon = ""

                report_sections.append(f"### {file_type} File: {logical_name}")
                report_sections.append(f"- **Size**: {size_mb:,.1f} MB")
                if used_mb is not None and used_percent is not None:
                    report_sections.append(
                        f"- **Used**: {used_mb:,.1f} MB ({used_percent:.1f}%) {status_icon}"
                    )
                else:
                    report_sections.append("- **Used**: N/A (FILESTREAM file)")
                report_sections.append(f"- **Max Size**: {max_size}")
                report_sections.append(f"- **Growth**: {growth}")
                report_sections.append("")
        else:
            report_sections.append("No file details available")

        # Growth Trend Analysis
        report_sections.append("## Storage Growth Analysis")
        growth_data = storage_data.get("growth_trend", {})
        if "error" in growth_data:
            report_sections.append(
                f"⚠️ **Growth analysis unavailable:** {growth_data['error']}"
            )
        elif growth_data.get("growth_analysis"):
            analysis = growth_data["growth_analysis"]
            total_growth = analysis.get("total_growth_mb", 0) or 0
            growth_percent = analysis.get("growth_percent", 0) or 0
            days_analyzed = analysis.get("days_analyzed", 0) or 0
            daily_growth = analysis.get("avg_daily_growth_mb", 0) or 0

            growth_icon = (
                "🔴" if daily_growth > 100 else "🟡" if daily_growth > 50 else "🟢"
            )

            report_sections.append(f"- **Analysis Period**: {days_analyzed} days")
            report_sections.append(
                f"- **Total Growth**: {total_growth:,.1f} MB ({growth_percent:.1f}%)"
            )
            report_sections.append(
                f"- **Daily Average Growth**: {daily_growth:,.1f} MB {growth_icon}"
            )

            # Growth projection
            if daily_growth > 0:
                days_to_double = (
                    (summary.get("total_database_size_mb", 0) / daily_growth)
                    if daily_growth > 0
                    else 0
                )
                report_sections.append(
                    f"- **Projected to Double**: {days_to_double:,.0f} days"
                )
        else:
            report_sections.append(
                "Growth analysis requires backup history (not available)"
            )
        report_sections.append("")

        # Top Tables by Space Usage
        report_sections.append(f"## Top {top_tables} Tables by Space Usage")
        table_usage = storage_data.get("table_usage", [])
        if table_usage:
            report_sections.append("")
            for i, table in enumerate(table_usage[:top_tables], 1):
                schema_name = table.get("schema_name", "unknown")
                table_name = table.get("table_name", "unknown")
                total_space = table.get("total_space_mb", 0)
                row_count = table.get("row_count", 0)
                index_type = table.get("index_type", "unknown")

                report_sections.append(f"### {i}. {schema_name}.{table_name}")
                report_sections.append(f"- **Total Space**: {total_space:,.1f} MB")
                report_sections.append(f"- **Row Count**: {row_count:,}")
                report_sections.append(f"- **Index Type**: {index_type}")
                report_sections.append("")
        else:
            report_sections.append("No table usage data available")

        # Azure Monitor Storage Metrics
        report_sections.append("## Azure Monitor Storage Metrics")
        metrics = storage_data.get("metrics", {})
        if "error" in metrics:
            report_sections.append(f"⚠️ **Metrics unavailable:** {metrics['error']}")
        else:
            metric_found = False
            for metric_name, metric_data in metrics.items():
                if metric_data:
                    metric_found = True
                    recent_values = metric_data[-5:]  # Last 5 data points
                    if recent_values:
                        avg_value = sum(
                            point.get("average", 0) or 0 for point in recent_values
                        ) / len(recent_values)
                        max_value = max(
                            point.get("maximum", 0) or 0 for point in recent_values
                        )

                        # Format based on metric type
                        if "percent" in metric_name:
                            report_sections.append(
                                f"- **{metric_name}**: Avg {avg_value:.1f}%, Max {max_value:.1f}%"
                            )
                        else:
                            report_sections.append(
                                f"- **{metric_name}**: Avg {avg_value:,.1f}, Max {max_value:,.1f}"
                            )

            if not metric_found:
                report_sections.append("No recent storage metric data available")

        # TempDB Usage
        tempdb_data = storage_data.get("tempdb", {})
        if tempdb_data and "error" not in tempdb_data:
            report_sections.append("")
            report_sections.append("## TempDB Usage")
            for metric_type, data in tempdb_data.items():
                if isinstance(data, dict):
                    used_percent = data.get("used_percent", 0) or 0
                    status_icon = (
                        "🔴"
                        if used_percent > 90
                        else "🟡"
                        if used_percent > 75
                        else "🟢"
                    )
                    report_sections.append(
                        f"- **{metric_type}**: {data.get('used_size_mb', 0) or 0:,.1f} MB / {data.get('total_size_mb', 0) or 0:,.1f} MB ({used_percent:.1f}%) {status_icon}"
                    )

        return "\n".join(report_sections)

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        try:
            hours_back = params.get("hours_back", 24)
            top_tables = params.get("top_tables", 20)

            db_config = self.toolset.database_config()

            # Create storage analysis API client
            api_client = self.toolset.api_client()
            storage_api = StorageAnalysisAPI(
                credential=api_client.credential,
                subscription_id=db_config.subscription_id,
            )

            # Gather storage data
            storage_data: Dict[str, Any] = {}

            # Get storage summary
            storage_data["summary"] = storage_api.get_storage_summary(
                db_config.server_name, db_config.database_name
            )

            # Get file details
            storage_data["file_details"] = storage_api.get_database_size_details(
                db_config.server_name, db_config.database_name
            )

            # Get table space usage
            storage_data["table_usage"] = storage_api.get_table_space_usage(
                db_config.server_name, db_config.database_name, top_tables
            )

            # Get growth trend
            storage_data["growth_trend"] = storage_api.get_storage_growth_trend(
                db_config.server_name, db_config.database_name
            )

            # Get Azure Monitor storage metrics
            storage_data["metrics"] = storage_api.get_storage_metrics(
                db_config.resource_group,
                db_config.server_name,
                db_config.database_name,
                hours_back,
            )

            # Get TempDB usage
            storage_data["tempdb"] = storage_api.get_tempdb_usage(
                db_config.server_name, db_config.database_name
            )

            # Build the formatted report
            report_text = self._build_storage_report(
                db_config, storage_data, hours_back, top_tables
            )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=report_text,
                params=params,
            )
        except Exception as e:
            error_msg = f"Failed to generate storage report: {str(e)}"
            logging.error(error_msg)
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=error_msg,
                params=params,
            )

    def get_parameterized_one_liner(self, params: Dict) -> str:
        db_config = self.toolset.database_config()
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Analyze Database Storage ({db_config.server_name}/{db_config.database_name})"

    @staticmethod
    def validate_config(
        api_client: AzureSQLAPIClient, database_config: AzureSQLDatabaseConfig
    ) -> Tuple[bool, str]:
        errors = []

        # Create storage analysis API client for validation
        storage_api = StorageAnalysisAPI(
            credential=api_client.credential,
            subscription_id=database_config.subscription_id,
        )

        # Test SQL database connection (storage queries)
        try:
            storage_api.get_storage_summary(
                database_config.server_name, database_config.database_name
            )
        except Exception as e:
            error_msg = str(e)
            if "authentication" in error_msg.lower() or "login" in error_msg.lower():
                errors.append(f"SQL database authentication failed: {error_msg}")
            elif (
                "permission" in error_msg.lower()
                or "authorization" in error_msg.lower()
            ):
                errors.append(f"SQL database permissions insufficient: {error_msg}")
            else:
                errors.append(f"SQL database connection failed: {error_msg}")

        # Test Azure Monitor API access (storage metrics)
        try:
            storage_api.get_storage_metrics(
                database_config.resource_group,
                database_config.server_name,
                database_config.database_name,
                1,  # Test with 1 hour
            )
        except Exception as e:
            error_msg = str(e)
            if (
                "authorization" in error_msg.lower()
                or "permission" in error_msg.lower()
            ):
                errors.append(f"Azure Monitor API access denied: {error_msg}")
            else:
                errors.append(f"Azure Monitor API connection failed: {error_msg}")

        if errors:
            return False, "\n".join(errors)
        return True, ""



import json
import logging
import os
from typing import Any, Optional, Dict, Tuple
from holmes.core.tools import (
    CallablePrerequisite,
    StructuredToolResult,
    Tool,
    ToolInvokeContext,
    ToolParameter,
    StructuredToolResultStatus,
    Toolset,
    ToolsetTag,
)
from holmes.plugins.toolsets.consts import (
    TOOLSET_CONFIG_MISSING_ERROR,
    STANDARD_END_DATETIME_TOOL_PARAM_DESCRIPTION,
)
from holmes.plugins.toolsets.datadog.datadog_api import (
    DatadogBaseConfig,
    DataDogRequestError,
    execute_datadog_http_request,
    get_headers,
    MAX_RETRY_COUNT_ON_RATE_LIMIT,
)
from holmes.plugins.toolsets.utils import (
    get_param_or_raise,
    process_timestamps_to_int,
    standard_start_datetime_tool_param_description,
    toolset_name_for_one_liner,
)
from holmes.plugins.toolsets.logging_utils.logging_api import (
    DEFAULT_TIME_SPAN_SECONDS,
    DEFAULT_LOG_LIMIT,
)

from datetime import datetime


class DatadogMetricsConfig(DatadogBaseConfig):
    default_limit: int = DEFAULT_LOG_LIMIT


class BaseDatadogMetricsTool(Tool):
    toolset: "DatadogMetricsToolset"


ACTIVE_METRICS_DEFAULT_LOOK_BACK_HOURS = 24
ACTIVE_METRICS_DEFAULT_TIME_SPAN_SECONDS = 24 * 60 * 60


class ListActiveMetrics(BaseDatadogMetricsTool):
    def __init__(self, toolset: "DatadogMetricsToolset"):
        super().__init__(
            name="list_active_datadog_metrics",
            description=f"[datadog/metrics toolset] List active metrics from Datadog for the last {ACTIVE_METRICS_DEFAULT_LOOK_BACK_HOURS} hours. This includes metrics that have actively reported data points, including from pods no longer in the cluster.",
            parameters={
                "from_time": ToolParameter(
                    description=f"Start time for listing metrics. Can be an RFC3339 formatted datetime (e.g. '2023-03-01T10:30:00Z') or a negative integer for relative seconds from now (e.g. -86400 for 24 hours ago). Defaults to {ACTIVE_METRICS_DEFAULT_LOOK_BACK_HOURS} hours ago",
                    type="string",
                    required=False,
                ),
                "host": ToolParameter(
                    description="Filter metrics by hostname",
                    type="string",
                    required=False,
                ),
                "tag_filter": ToolParameter(
                    description="Filter metrics by tags in the format tag:value.",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.dd_config:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=TOOLSET_CONFIG_MISSING_ERROR,
                params=params,
            )

        url = None
        query_params = None

        try:
            from_time_str = params.get("from_time")

            from_time, _ = process_timestamps_to_int(
                start=from_time_str,
                end=None,
                default_time_span_seconds=ACTIVE_METRICS_DEFAULT_TIME_SPAN_SECONDS,
            )

            url = f"{self.toolset.dd_config.site_api_url}/api/v1/metrics"
            headers = get_headers(self.toolset.dd_config)

            query_params = {
                "from": from_time,
            }

            if params.get("host"):
                query_params["host"] = params["host"]

            if params.get("tag_filter"):
                query_params["tag_filter"] = params["tag_filter"]

            data = execute_datadog_http_request(
                url=url,
                headers=headers,
                timeout=self.toolset.dd_config.request_timeout,
                method="GET",
                payload_or_params=query_params,
            )

            metrics = data.get("metrics", [])
            if not metrics:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    data="Your filter returned no metrics. Change your filter and try again",
                    params=params,
                )

            output = ["Metric Name"]
            output.append("-" * 50)

            for metric in sorted(metrics):
                output.append(metric)

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data="\n".join(output),
                params=params,
            )

        except DataDogRequestError as e:
            logging.exception(e, exc_info=True)

            if e.status_code == 429:
                error_msg = f"Datadog API rate limit exceeded. Failed after {MAX_RETRY_COUNT_ON_RATE_LIMIT} retry attempts."
            elif e.status_code == 403:
                error_msg = (
                    f"Permission denied. Ensure your Datadog Application Key has the 'metrics_read' "
                    f"and 'timeseries_query' permissions. Error: {str(e)}"
                )
            else:
                # Include full API error details for better debugging
                error_msg = (
                    f"Datadog API error (status {e.status_code}): {e.response_text}"
                )
                if params:
                    # ListActiveMetrics parameters: from_time, host, tag_filter
                    if params.get("host"):
                        error_msg += f"\nHost filter: {params.get('host')}"
                    if params.get("tag_filter"):
                        error_msg += f"\nTag filter: {params.get('tag_filter')}"

                    from_time_param = params.get("from_time")
                    if from_time_param:
                        time_desc = from_time_param
                    else:
                        time_desc = f"default (last {ACTIVE_METRICS_DEFAULT_LOOK_BACK_HOURS} hours)"
                    error_msg += f"\nTime range: {time_desc}"

                    # Note: We cannot generate a Datadog Metrics Explorer URL for ListActiveMetrics
                    # because the Metrics Explorer requires a specific metric query,
                    # while ListActiveMetrics just lists available metrics without querying any specific one

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=error_msg,
                params=params,
                invocation=json.dumps({"url": url, "params": query_params})
                if url and query_params
                else None,
            )

        except Exception as e:
            logging.exception(
                f"Failed to query Datadog metrics for params: {params}", exc_info=True
            )
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Exception while querying Datadog: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        filters = []
        if params.get("host"):
            filters.append(f"host={params['host']}")
        if params.get("tag_filter"):
            filters.append(f"tag_filter={params['tag_filter']}")
        filter_str = f"{', '.join(filters)}" if filters else "all"
        return f"{toolset_name_for_one_liner(self.toolset.name)}: List Active Metrics ({filter_str})"


class QueryMetrics(BaseDatadogMetricsTool):
    def __init__(self, toolset: "DatadogMetricsToolset"):
        super().__init__(
            name="query_datadog_metrics",
            description="[datadog/metrics toolset] Query timeseries data from Datadog for a specific metric, including historical data for pods no longer in the cluster",
            parameters={
                "query": ToolParameter(
                    description="The metric query string (e.g., 'system.cpu.user{host:myhost}')",
                    type="string",
                    required=True,
                ),
                "from_time": ToolParameter(
                    description=standard_start_datetime_tool_param_description(
                        DEFAULT_TIME_SPAN_SECONDS
                    ),
                    type="string",
                    required=False,
                ),
                "to_time": ToolParameter(
                    description=STANDARD_END_DATETIME_TOOL_PARAM_DESCRIPTION,
                    type="string",
                    required=False,
                ),
                "description": ToolParameter(
                    description="Describes the query",
                    type="string",
                    required=True,
                ),
                "output_type": ToolParameter(
                    description="Specifies how to interpret the Datadog result. Use 'Plain' for raw values, 'Bytes' to format byte values, 'Percentage' to scale 0–1 values into 0–100%, or 'CPUUsage' to convert values to cores (e.g., 500 becomes 500m, 2000 becomes 2).",
                    type="string",
                    required=False,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.dd_config:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=TOOLSET_CONFIG_MISSING_ERROR,
                params=params,
            )

        url = None
        query_params = None

        try:
            query = get_param_or_raise(params, "query")

            (from_time, to_time) = process_timestamps_to_int(
                start=params.get("from_time"),
                end=params.get("to_time"),
                default_time_span_seconds=DEFAULT_TIME_SPAN_SECONDS,
            )

            url = f"{self.toolset.dd_config.site_api_url}/api/v1/query"
            headers = get_headers(self.toolset.dd_config)

            query_params = {
                "query": query,
                "from": from_time,
                "to": to_time,
            }

            data = execute_datadog_http_request(
                url=url,
                headers=headers,
                timeout=self.toolset.dd_config.request_timeout,
                method="GET",
                payload_or_params=query_params,
            )

            series = data.get("series", [])
            description = params.get("description", "")
            output_type = params.get("output_type", "Plain")

            if not series:
                # Include detailed context in error message
                from_time_param = params.get("from_time")
                to_time_param = params.get("to_time")

                if from_time_param:
                    from_desc = from_time_param
                else:
                    from_desc = (
                        f"default (last {DEFAULT_TIME_SPAN_SECONDS // 86400} days)"
                    )

                to_desc = to_time_param or "now"

                error_msg = (
                    f"The query returned no data.\n"
                    f"Query: {params.get('query', 'not specified')}\n"
                    f"Time range: {from_desc} to {to_desc}\n"
                    f"Please check your query syntax and ensure data exists for this time range."
                )

                return StructuredToolResult(
                    status=StructuredToolResultStatus.NO_DATA,
                    error=error_msg,
                    params=params,
                )

            # Transform Datadog series data to match Prometheus format
            prometheus_result = []
            for serie in series:
                # Extract metric info from Datadog series
                metric_info = {}
                if "metric" in serie:
                    metric_info["__name__"] = serie["metric"]

                # Add other fields from scope/tag_set if available
                if "scope" in serie and serie["scope"]:
                    # Parse scope like "pod_name:robusta-runner-78599b764d-f847h" into labels
                    scope_parts = serie["scope"].split(",")
                    for part in scope_parts:
                        if ":" in part:
                            key, value = part.split(":", 1)
                            metric_info[key.strip()] = value.strip()

                # Transform pointlist to values format (timestamp, value as strings)
                values = []
                if "pointlist" in serie:
                    for point in serie["pointlist"]:
                        if len(point) >= 2:
                            # Convert timestamp from milliseconds to seconds, format as string
                            timestamp = int(point[0] / 1000)
                            value = str(point[1])
                            values.append([timestamp, value])

                prometheus_result.append({"metric": metric_info, "values": values})

            # Convert timestamps to RFC3339 format for start/end
            start_rfc = datetime.fromtimestamp(from_time).strftime("%Y-%m-%dT%H:%M:%SZ")
            end_rfc = datetime.fromtimestamp(to_time).strftime("%Y-%m-%dT%H:%M:%SZ")

            # Create response matching Prometheus format
            response_data = {
                "status": "success",
                "error_message": None,
                "tool_name": self.name,
                "description": description,
                "query": query,
                "start": start_rfc,
                "end": end_rfc,
                "step": 60,  # Default step, Datadog doesn't provide this directly
                "output_type": output_type,
                "data": {"resultType": "matrix", "result": prometheus_result},
            }

            data_str = json.dumps(response_data, indent=2)
            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data_str,
                params=params,
            )

        except DataDogRequestError as e:
            logging.exception(e, exc_info=True)

            if e.status_code == 429:
                error_msg = f"Datadog API rate limit exceeded. Failed after {MAX_RETRY_COUNT_ON_RATE_LIMIT} retry attempts."
            elif e.status_code == 403:
                error_msg = (
                    f"Permission denied. Ensure your Datadog Application Key has the 'metrics_read' "
                    f"and 'timeseries_query' permissions. Error: {str(e)}"
                )
            else:
                # Include full API error details for better debugging
                error_msg = (
                    f"Datadog API error (status {e.status_code}): {e.response_text}"
                )
                if params:
                    error_msg += f"\nQuery: {params.get('query', 'not specified')}"

                    from_time_param = params.get("from_time")
                    to_time_param = params.get("to_time")

                    if from_time_param:
                        from_desc = from_time_param
                    else:
                        from_desc = (
                            f"default (last {DEFAULT_TIME_SPAN_SECONDS // 86400} days)"
                        )

                    to_desc = to_time_param or "now"
                    error_msg += f"\nTime range: {from_desc} to {to_desc}"

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=error_msg,
                params=params,
                invocation=json.dumps({"url": url, "params": query_params})
                if url and query_params
                else None,
            )

        except Exception as e:
            logging.exception(
                f"Failed to query Datadog metrics for params: {params}", exc_info=True
            )

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Exception while querying Datadog: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        description = params.get("description", "")
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Query Metrics ({description})"


class QueryMetricsMetadata(BaseDatadogMetricsTool):
    def __init__(self, toolset: "DatadogMetricsToolset"):
        super().__init__(
            name="get_datadog_metric_metadata",
            description="[datadog/metrics toolset] Get metadata about one or more metrics including their type, description, unit, and other properties",
            parameters={
                "metric_names": ToolParameter(
                    description="Comma-separated list of metric names to get metadata for (e.g., 'system.cpu.user, system.mem.used')",
                    type="string",
                    required=True,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.dd_config:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=TOOLSET_CONFIG_MISSING_ERROR,
                params=params,
            )

        try:
            metric_names_str = get_param_or_raise(params, "metric_names")

            metric_names = [
                name.strip()
                for name in metric_names_str.split(",")
                if name.strip()  # Filter out empty strings
            ]

            if not metric_names:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error="metric_names cannot be empty",
                    params=params,
                )

            headers = get_headers(self.toolset.dd_config)

            results = {}
            errors = {}

            for metric_name in metric_names:
                try:
                    url = f"{self.toolset.dd_config.site_api_url}/api/v1/metrics/{metric_name}"

                    data = execute_datadog_http_request(
                        url=url,
                        headers=headers,
                        payload_or_params={},
                        timeout=self.toolset.dd_config.request_timeout,
                        method="GET",
                    )

                    results[metric_name] = data

                except DataDogRequestError as e:
                    if e.status_code == 404:
                        errors[metric_name] = "Metric not found"
                    elif e.status_code == 429:
                        errors[metric_name] = (
                            f"Datadog API rate limit exceeded. Failed after {MAX_RETRY_COUNT_ON_RATE_LIMIT} retry attempts."
                        )
                    else:
                        errors[metric_name] = f"Error {e.status_code}: {str(e)}"
                except Exception as e:
                    errors[metric_name] = f"Exception: {str(e)}"

            response_data = {
                "metrics_metadata": results,
                "errors": errors,
                "total_requested": len(metric_names),
                "successful": len(results),
                "failed": len(errors),
            }

            if not results and errors:
                return StructuredToolResult(
                    status=StructuredToolResultStatus.ERROR,
                    error="Failed to retrieve metadata for all metrics",
                    data=json.dumps(response_data, indent=2),
                    params=params,
                )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=json.dumps(response_data, indent=2),
                params=params,
            )

        except Exception as e:
            logging.exception(
                f"Failed to query Datadog metric metadata for params: {params}",
                exc_info=True,
            )

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Exception while querying Datadog: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        metric_names = params.get("metric_names", [])
        if isinstance(metric_names, list):
            if len(metric_names) == 1:
                return f"Get Metric Metadata ({metric_names[0]})"
            elif len(metric_names) > 1:
                return f"{toolset_name_for_one_liner(self.toolset.name)}: Get Datadog metric metadata for {len(metric_names)} metrics"
        return f"{toolset_name_for_one_liner(self.toolset.name)}: Get Datadog metric metadata"


class ListMetricTags(BaseDatadogMetricsTool):
    def __init__(self, toolset: "DatadogMetricsToolset"):
        super().__init__(
            name="list_datadog_metric_tags",
            description="[datadog/metrics toolset] List all available tags and aggregations for a specific metric. This helps in building queries by showing what dimensions are available for filtering.",
            parameters={
                "metric_name": ToolParameter(
                    description="The name of the metric to get tags for (e.g., 'system.cpu.user', 'container.memory.usage')",
                    type="string",
                    required=True,
                ),
            },
            toolset=toolset,
        )

    def _invoke(self, params: dict, context: ToolInvokeContext) -> StructuredToolResult:
        if not self.toolset.dd_config:
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=TOOLSET_CONFIG_MISSING_ERROR,
                params=params,
            )

        url = None
        query_params = None

        try:
            metric_name = get_param_or_raise(params, "metric_name")

            url = f"{self.toolset.dd_config.site_api_url}/api/v2/metrics/{metric_name}/active-configurations"
            headers = get_headers(self.toolset.dd_config)

            data = execute_datadog_http_request(
                url=url,
                headers=headers,
                timeout=self.toolset.dd_config.request_timeout,
                method="GET",
                payload_or_params={},
            )

            return StructuredToolResult(
                status=StructuredToolResultStatus.SUCCESS,
                data=data,
                params=params,
            )

        except DataDogRequestError as e:
            logging.exception(e, exc_info=True)

            if e.status_code == 404:
                error_msg = f"Metric '{params.get('metric_name', 'unknown')}' not found. Please check the metric name."
            elif e.status_code == 429:
                error_msg = f"Datadog API rate limit exceeded. Failed after {MAX_RETRY_COUNT_ON_RATE_LIMIT} retry attempts."
            elif e.status_code == 403:
                error_msg = (
                    f"Permission denied. Ensure your Datadog Application Key has the 'metrics_read' "
                    f"permissions. Error: {str(e)}"
                )
            else:
                # Include full API error details for better debugging
                error_msg = (
                    f"Datadog API error (status {e.status_code}): {e.response_text}"
                )
                if params:
                    error_msg += (
                        f"\nMetric name: {params.get('metric_name', 'not specified')}"
                    )

            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=error_msg,
                params=params,
                invocation=json.dumps({"url": url, "params": query_params})
                if url and query_params
                else None,
            )

        except Exception as e:
            logging.exception(
                f"Failed to query Datadog metric tags for params: {params}",
                exc_info=True,
            )
            return StructuredToolResult(
                status=StructuredToolResultStatus.ERROR,
                error=f"Exception while querying Datadog: {str(e)}",
                params=params,
            )

    def get_parameterized_one_liner(self, params) -> str:
        metric_name = params.get("metric_name", "<no metric>")
        return f"List available tags for Datadog metric: {metric_name}"


class DatadogMetricsToolset(Toolset):
    dd_config: Optional[DatadogMetricsConfig] = None

    def __init__(self):
        super().__init__(
            name="datadog/metrics",
            description="Toolset for fetching metrics and metadata from Datadog, including historical data for pods no longer in the cluster",
            docs_url="https://holmesgpt.dev/data-sources/builtin-toolsets/datadog/",
            icon_url="https://imgix.datadoghq.com//img/about/presskit/DDlogo.jpg",
            prerequisites=[CallablePrerequisite(callable=self.prerequisites_callable)],
            tools=[
                ListActiveMetrics(toolset=self),
                QueryMetrics(toolset=self),
                QueryMetricsMetadata(toolset=self),
                ListMetricTags(toolset=self),
            ],
            tags=[ToolsetTag.CORE],
        )
        self._reload_instructions()

    def _perform_healthcheck(self, dd_config: DatadogMetricsConfig) -> Tuple[bool, str]:
        try:
            logging.debug("Performing Datadog metrics configuration healthcheck...")

            url = f"{dd_config.site_api_url}/api/v1/validate"
            headers = get_headers(dd_config)

            data = execute_datadog_http_request(
                url=url,
                headers=headers,
                payload_or_params={},
                timeout=dd_config.request_timeout,
                method="GET",
            )

            if data.get("valid", False):
                logging.info("Datadog metrics healthcheck completed successfully")
                return True, ""
            else:
                error_msg = "Datadog API key validation failed"
                logging.error(f"Datadog metrics healthcheck failed: {error_msg}")
                return False, f"Datadog metrics healthcheck failed: {error_msg}"

        except Exception as e:
            logging.exception("Failed during Datadog metrics healthcheck")
            return False, f"Healthcheck failed with exception: {str(e)}"

    def prerequisites_callable(self, config: dict[str, Any]) -> Tuple[bool, str]:
        if not config:
            return (
                False,
                "Missing config for dd_api_key, dd_app_key, or site_api_url. For details: https://holmesgpt.dev/data-sources/builtin-toolsets/datadog/",
            )

        try:
            dd_config = DatadogMetricsConfig(**config)
            self.dd_config = dd_config

            success, error_msg = self._perform_healthcheck(dd_config)
            return success, error_msg

        except Exception as e:
            logging.exception("Failed to set up Datadog metrics toolset")
            return (False, f"Failed to parse Datadog configuration: {str(e)}")

    def get_example_config(self) -> Dict[str, Any]:
        return {
            "dd_api_key": "your-datadog-api-key",
            "dd_app_key": "your-datadog-application-key",
            "site_api_url": "https://api.datadoghq.com",
            "default_limit": 1000,
            "request_timeout": 60,
        }

    def _reload_instructions(self):
        """Load Datadog metrics specific troubleshooting instructions."""
        template_file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__), "datadog_metrics_instructions.jinja2"
            )
        )
        self._load_llm_instructions(jinja_template=f"file://{template_file_path}")
